# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#  under the License.


import contextlib
import cProfile as profiler
import gc
import sys
import time
try:
    import pstats
except Exception:
    # Don't want to force pstats into the venv if it's not always used
    pass

from neutron.api.v2 import attributes
from neutron_lib import exceptions as n_exc
from oslo_log import log as logging
import webob

LOG = logging.getLogger(__name__)


def filter_body(context, body, admin_only=None, always_filter=None):
    if not context.is_admin and admin_only:
        for attr in admin_only:
            pop_param(body, attr)
    if always_filter:
        for attr in always_filter:
            pop_param(body, attr)


def attr_specified(param):
    return param is not attributes.ATTR_NOT_SPECIFIED


def timed(fn):
    def _wrapped(*args, **kwargs):
        began = time.time()
        res = fn(*args, **kwargs)
        elapsed = time.time() - began
        LOG.info("Time for %s = %s" % (fn, elapsed))
        return res
    return _wrapped


def profile(output):
    def _inner(fn):
        def _wrapped(*args, **kw):
            result = _profile(output, fn, *args, **kw)
            # uncomment this to see who's calling what
            # stats.print_callers()
            return result
        return _wrapped
    return _inner


def live_profile(fn):
    def _wrapped(*args, **kw):
        elapsed, stat_loader, result = _live_profile(fn, *args, **kw)
        stats = stat_loader()
        stats.sort_stats('cumulative')
        stats.print_stats()
        # uncomment this to see who's calling what
        # stats.print_callers()
        return result
    return _wrapped


def exc_wrapper(func):
    def wrapped(*args, **kwargs):
        e = None
        try:
            return func(*args, **kwargs)
        except n_exc.NotFound as e:
            raise webob.exc.HTTPNotFound(e)
        except n_exc.Conflict as e:
            raise webob.exc.HTTPConflict(e)
        except n_exc.BadRequest as e:
            raise webob.exc.HTTPBadRequest(e)
        except Exception as e:
            raise webob.exc.HTTPInternalServerError(e)
        finally:
            if not e:
                LOG.exception(str(e))
    return wrapped


def _profile(filename, fn, *args, **kw):
    gc.collect()

    profiler.runctx('result = fn(*args, **kw)', globals(), locals(),
                    filename=filename)

    return locals()['result']


def _live_profile(fn, *args, **kw):
    load_stats = lambda: pstats.Stats()
    gc.collect()

    began = time.time()
    profiler.runctx('result = fn(*args, **kw)', globals(), locals())
    ended = time.time()

    return ended - began, load_stats, locals()['result']


def pop_param(attrs, param, default=None):
    val = attrs.pop(param, default)
    if attr_specified(val):
        return val
    return default


class Command(object):
    def __init__(self, func):
        self.func = func
        self.result = None
        self.called = False
        self.is_rollback = False

    def __call__(self, *args, **kwargs):
        self.called = True
        kwargs['rollback'] = self.is_rollback
        self.result = self.func(*args, **kwargs)
        return self.result


class CommandManager(object):
    def __init__(self):
        self.do_commands = []
        self.undo_commands = []

    @contextlib.contextmanager
    def execute(self, exc=None):
        try:
            yield self
        except Exception:
            exc_info = sys.exc_info()
            LOG.exception("Exception in transaction", exc_info=exc_info)
            self.rollback()
            raise exc_info[1]

    def do(self, func):
        cmd = Command(func)
        self.do_commands.append(cmd)
        return cmd

    def undo(self, func):
        cmd = Command(func)
        cmd.is_rollback = True
        self.undo_commands.append(cmd)
        return cmd

    def rollback(self):
        do_commands = reversed(self.do_commands)
        for cmd in reversed(self.undo_commands):
            do = do_commands.next()
            if not do.called:
                continue
            try:
                cmd(do.result)
            except Exception:
                LOG.exception("Rollback failed and wasn't caught!")


class retry_loop(object):
    def __init__(self, retry_times, delay=0, backoff=1):
        self._retry_times = retry_times
        self._delay = delay
        self._backoff = backoff

    def __call__(self, f):
        def wrapped_f(*args, **kwargs):
            level = self._retry_times
            current_delay = self._delay
            while level > 0:
                try:
                    return f(*args, **kwargs)
                except Exception:
                    level = level - 1
                    if level > 0:
                        if current_delay > 0:
                            time.sleep(current_delay)
                            if self._backoff > 0:
                                current_delay *= self._backoff

                        LOG.debug("Retrying `%s` %d more times...",
                                  f.func_name, level)
                    else:
                        raise
        return wrapped_f


def pretty_kwargs(**kwargs):
    kwargs_str = ', '.join("%s=%s" % (k, v) for k, v in kwargs.items())
    return kwargs_str



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#  under the License.

import contextlib

import mock
from neutron.extensions import securitygroup as sg_ext
from neutron_lib import exceptions as n_exc
from oslo_config import cfg

from quark.db import models
from quark import exceptions as q_exc
from quark.plugin_modules import security_groups
from quark import protocols
from quark.tests import test_quark_plugin


class TestQuarkGetSecurityGroups(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, security_groups):
        def _make_rules(rules):
            rule_mods = []
            for rule in rules:
                r = models.SecurityGroupRule()
                r.update(rule)
                rule_mods.append(r)
            return rule_mods

        if isinstance(security_groups, list):
            for sg in security_groups:
                sg["rules"] = _make_rules(sg["rules"])
        elif security_groups:
            security_groups["rules"] = _make_rules(security_groups["rules"])

        with mock.patch("quark.db.api.security_group_find") as db_find:
            db_find.return_value = security_groups
            yield

    def test_get_security_groups_list(self):
        v4_ethertype = protocols.ETHERTYPES["IPv4"]
        rule = {"ethertype": v4_ethertype,
                "protocol": protocols.PROTOCOL_MAP[v4_ethertype]["tcp"]}
        group = {"name": "foo", "description": "bar",
                 "tenant_id": self.context.tenant_id, "rules": [rule]}
        with self._stubs([group]):
            result = self.plugin.get_security_groups(self.context, {})
            group = result[0]
            self.assertEqual("fake", group["tenant_id"])
            self.assertEqual("foo", group["name"])
            self.assertEqual("bar", group["description"])
            rule = group["security_group_rules"][0]
            self.assertEqual("IPv4", rule["ethertype"])
            self.assertEqual("TCP", rule["protocol"])

    def test_get_security_group(self):
        v4_ethertype = protocols.ETHERTYPES["IPv4"]
        rule = {"ethertype": v4_ethertype,
                "protocol": protocols.PROTOCOL_MAP[v4_ethertype]["tcp"]}

        group = {"name": "foo", "description": "bar",
                 "tenant_id": self.context.tenant_id, "rules": [rule]}
        with self._stubs(group):
            result = self.plugin.get_security_group(self.context, 1)
            self.assertEqual("fake", result["tenant_id"])
            self.assertEqual("foo", result["name"])
            self.assertEqual("bar", result["description"])
            rule = result["security_group_rules"][0]
            self.assertEqual("IPv4", rule["ethertype"])
            self.assertEqual("TCP", rule["protocol"])

    def test_get_security_group_group_not_found_fails(self):
        with self._stubs(security_groups=None):
            with self.assertRaises(sg_ext.SecurityGroupNotFound):
                self.plugin.get_security_group(self.context, 1)


class TestQuarkGetSecurityGroupRules(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, security_rules):
        if isinstance(security_rules, list):
            rules = []
            for rule in security_rules:
                r = models.SecurityGroupRule()
                r.update(rule)
                rules.append(r)
        elif security_rules is not None:
            rules = models.SecurityGroupRule()
            rules.update(security_rules)
        with mock.patch("quark.db.api.security_group_rule_find") as db_find:
            db_find.return_value = security_rules
            yield

    def test_get_security_group_rules(self):
        rule = {"id": 1, "remote_group_id": 2, "direction": "ingress",
                "port_range_min": 80, "port_range_max": 100,
                "remote_ip_prefix": None,
                "ethertype": protocols.translate_ethertype("IPv4"),
                "tenant_id": "foo", "protocol": "UDP", "group_id": 1}
        expected = rule.copy()
        expected["security_group_id"] = expected.pop("group_id")
        expected["ethertype"] = "IPv4"

        with self._stubs([rule]):
            resp = self.plugin.get_security_group_rules(self.context, {})
            for key in expected.keys():
                self.assertTrue(key in resp[0])
                self.assertEqual(resp[0][key], expected[key])

    def test_get_security_group_rule(self):
        rule = {"id": 1, "remote_group_id": 2, "direction": "ingress",
                "port_range_min": 80, "port_range_max": 100,
                "remote_ip_prefix": None,
                "ethertype": protocols.translate_ethertype("IPv4"),
                "tenant_id": "foo", "protocol": "UDP", "group_id": 1}
        expected = rule.copy()
        expected["security_group_id"] = expected.pop("group_id")
        expected["ethertype"] = "IPv4"

        with self._stubs(rule):
            resp = self.plugin.get_security_group_rule(self.context, 1)
            for key in expected.keys():
                self.assertTrue(key in resp)
                self.assertEqual(resp[key], expected[key])

    def test_get_security_group_rule_not_found(self):
        with self._stubs(None):
            with self.assertRaises(sg_ext.SecurityGroupRuleNotFound):
                self.plugin.get_security_group_rule(self.context, 1)


class TestQuarkUpdateSecurityGroup(test_quark_plugin.TestQuarkPlugin):
    def test_update_security_group(self):
        v4_ethertype = protocols.ETHERTYPES["IPv4"]
        rule_dict = {"ethertype": v4_ethertype,
                     "protocol": protocols.PROTOCOL_MAP[v4_ethertype]["tcp"]}
        rule = models.SecurityGroupRule()
        rule.update(rule_dict)
        group = {"name": "foo", "description": "bar",
                 "tenant_id": self.context.tenant_id, "rules": [rule]}
        updated_group = group.copy()
        updated_group["name"] = "bar"

        with contextlib.nested(
                mock.patch("quark.db.api.security_group_find"),
                mock.patch("quark.db.api.security_group_update"),
        ) as (db_find, db_update):
            db_find.return_value = group
            db_update.return_value = updated_group
            update = dict(security_group=dict(name="bar"))
            resp = self.plugin.update_security_group(self.context, 1, update)
            self.assertEqual(resp["name"], updated_group["name"])

    def test_update_security_group_with_deault_security_group_id(self):
        with self.assertRaises(sg_ext.SecurityGroupCannotUpdateDefault):
            self.plugin.update_security_group(self.context,
                                              security_groups.DEFAULT_SG_UUID,
                                              None)


class TestQuarkCreateSecurityGroup(test_quark_plugin.TestQuarkPlugin):
    def setUp(self, *args, **kwargs):
        super(TestQuarkCreateSecurityGroup, self).setUp(*args, **kwargs)
        cfg.CONF.set_override('quota_security_group', 1, 'QUOTAS')

    @contextlib.contextmanager
    def _stubs(self, security_group, other=0):
        dbgroup = models.SecurityGroup()
        dbgroup.update(security_group)

        with contextlib.nested(
                mock.patch("quark.db.api.security_group_find"),
                mock.patch("quark.db.api.security_group_create"),
        ) as (db_find, db_create):
            db_find.return_value.count.return_value = other
            db_create.return_value = dbgroup
            yield db_create

    def test_create_security_group(self):
        group = {'name': 'foo', 'description': 'bar',
                 'tenant_id': self.context.tenant_id}
        expected = {'name': 'foo', 'description': 'bar',
                    'tenant_id': self.context.tenant_id,
                    'security_group_rules': []}
        with self._stubs(group) as group_create:
            result = self.plugin.create_security_group(
                self.context, {'security_group': group})
            self.assertTrue(group_create.called)
            for key in expected.keys():
                self.assertEqual(result[key], expected[key])

    def test_create_default_security_group(self):
        group = {'name': 'default', 'description': 'bar',
                 'tenant_id': self.context.tenant_id}
        with self._stubs(group) as group_create:
            with self.assertRaises(sg_ext.SecurityGroupDefaultAlreadyExists):
                self.plugin.create_security_group(
                    self.context, {'security_group': group})
                self.assertTrue(group_create.called)

    def test_create_security_group_name_too_long(self):
        group = {'name': 'a' * 256, 'description': 'bar',
                 'tenant_id': self.context.tenant_id}
        with self._stubs(group):
            with self.assertRaises(n_exc.InvalidInput):
                self.plugin.create_security_group(
                    self.context, {'security_group': group})

    def test_create_security_group_description(self):
        group = {'name': 'foo', 'description': 'b' * 256,
                 'tenant_id': self.context.tenant_id}
        with self._stubs(group):
            with self.assertRaises(n_exc.InvalidInput):
                self.plugin.create_security_group(
                    self.context, {'security_group': group})


class TestQuarkDeleteSecurityGroup(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, security_group=None):
        dbgroup = None
        if security_group:
            dbgroup = models.SecurityGroup()
            dbgroup.update(security_group)

        with contextlib.nested(
            mock.patch("quark.db.api.security_group_find"),
            mock.patch("quark.db.api.security_group_delete"),
        ) as (group_find, db_group_delete):
            group_find.return_value = dbgroup
            db_group_delete.return_value = dbgroup
            yield db_group_delete

    def test_delete_security_group(self):
        group = {'name': 'foo', 'description': 'bar', 'id': 1,
                 'tenant_id': self.context.tenant_id}
        with self._stubs(group) as (db_delete):
            self.plugin.delete_security_group(self.context, 1)
            self.assertTrue(db_delete.called)

    def test_delete_default_security_group(self):
        group = {'name': 'default', 'id': 1,
                 'tenant_id': self.context.tenant_id}
        with self._stubs(group):
            with self.assertRaises(sg_ext.SecurityGroupCannotRemoveDefault):
                self.plugin.delete_security_group(self.context, 1)

    def test_delete_security_group_with_ports(self):
        port = models.Port()
        group = {'name': 'foo', 'description': 'bar', 'id': 1,
                 'tenant_id': self.context.tenant_id, 'ports': [port]}
        with self._stubs(group):
            with self.assertRaises(sg_ext.SecurityGroupInUse):
                self.plugin.delete_security_group(self.context, 1)

    def test_delete_security_group_not_found(self):
        with self._stubs():
            with self.assertRaises(sg_ext.SecurityGroupNotFound):
                self.plugin.delete_security_group(self.context, 1)


class TestQuarkCreateSecurityGroupRule(test_quark_plugin.TestQuarkPlugin):
    def setUp(self, *args, **kwargs):
        super(TestQuarkCreateSecurityGroupRule, self).setUp(*args, **kwargs)
        cfg.CONF.set_override('quota_security_group_rule', 1, 'QUOTAS')
        cfg.CONF.set_override('quota_security_rules_per_group', 1, 'QUOTAS')
        self.rule = {'id': 1, 'ethertype': 'IPv4',
                     'security_group_id': 1, 'group': {'id': 1},
                     'protocol': None, 'port_range_min': None,
                     'port_range_max': None}
        self.expected = {
            'remote_group_id': None,
            'direction': None,
            'port_range_min': None,
            'port_range_max': None,
            'remote_ip_prefix': None,
            'ethertype': 'IPv4',
            'tenant_id': None,
            'protocol': None,
            'security_group_id': 1}

    @contextlib.contextmanager
    def _stubs(self, rule, group, limit_raise=False):
        dbgroup = None
        if group:
            dbgroup = models.SecurityGroup()
            dbgroup.update(group)

        def _create_rule(context, **rule):
            dbrule = models.SecurityGroupRule()
            dbrule.update(rule)
            dbrule["group_id"] = rule['security_group_id']
            return dbrule

        with contextlib.nested(
            mock.patch("quark.db.api.security_group_find"),
            mock.patch("quark.db.api.security_group_rule_find"),
            mock.patch("quark.db.api.security_group_rule_create"),
            mock.patch("quark.protocols.human_readable_protocol"),
            mock.patch("neutron.quota.QuotaEngine.limit_check")
        ) as (group_find, rule_find, rule_create, human, limit_check):
            group_find.return_value = dbgroup
            rule_find.return_value.count.return_value = group.get(
                'port_rules', None) if group else 0

            rule_create.side_effect = _create_rule
            human.return_value = rule["protocol"]
            if limit_raise:
                limit_check.side_effect = n_exc.OverQuota
            yield rule_create

    def _test_create_security_rule(self, limit_raise=False, **ruleset):
        ruleset['tenant_id'] = self.context.tenant_id
        rule = dict(self.rule, **ruleset)
        group = rule.pop('group')
        expected = dict(self.expected, **ruleset)
        expected.pop('group', None)
        hax = {'security_group_rule': rule}
        cfg.CONF.set_override('environment_capabilities',
                              ['security_groups', 'egress'], 'QUARK')
        with self._stubs(rule, group, limit_raise) as rule_create:
            result = self.plugin.create_security_group_rule(self.context, hax)
            self.assertTrue(rule_create.called)
            for key in expected.keys():
                self.assertEqual(expected[key], result[key])
        cfg.CONF.clear_override('environment_capabilities', 'QUARK')

    def test_create_security_rule_IPv6(self):
        self._test_create_security_rule(ethertype='IPv6')

    def test_create_security_rule_TCP(self):
        self._test_create_security_rule(protocol=6)

    def test_create_security_rule_ICMP(self):
        self._test_create_security_rule(protocol=1)

    def test_create_security_rule_ICMP_with_type(self):
        self._test_create_security_rule(protocol=1, port_range_min=1)

    def test_create_security_rule_ICMP_with_type_and_code(self):
        self._test_create_security_rule(protocol=1, port_range_min=1,
                                        port_range_max=2)

    def test_create_security_rule_ICMP_code_no_type_raises(self):
        with self.assertRaises(sg_ext.SecurityGroupMissingIcmpType):
            self._test_create_security_rule(protocol=1, port_range_max=2)

    def test_create_security_rule_ICMP_type_under_range_raises(self):
        with self.assertRaises(sg_ext.SecurityGroupInvalidIcmpValue):
            self._test_create_security_rule(protocol=1, port_range_min=-1,
                                            port_range_max=1)

    def test_create_security_rule_ICMP_type_over_range_raises(self):
        with self.assertRaises(sg_ext.SecurityGroupInvalidIcmpValue):
            self._test_create_security_rule(protocol=1, port_range_min=256,
                                            port_range_max=1)

    def test_create_security_rule_ICMP_code_under_range_raises(self):
        with self.assertRaises(sg_ext.SecurityGroupInvalidIcmpValue):
            self._test_create_security_rule(protocol=1, port_range_min=1,
                                            port_range_max=-1)

    def test_create_security_rule_ICMP_code_over_range_raises(self):
        with self.assertRaises(sg_ext.SecurityGroupInvalidIcmpValue):
            self._test_create_security_rule(protocol=1, port_range_min=1,
                                            port_range_max=256)

    def test_create_security_rule_remote_ip(self):
        self._test_create_security_rule(remote_ip_prefix='192.168.0.1')

    def test_create_security_rule_remote_group(self):
        with self.assertRaises(n_exc.InvalidInput):
            self._test_create_security_rule(remote_group_id=2)

    def test_create_security_rule_port_range_invalid_ranges_fails(self):
        with self.assertRaises(n_exc.InvalidInput):
            self._test_create_security_rule(protocol=6, port_range_min=0)

    def test_create_security_rule_min_under_port_min(self):
        with self.assertRaises(n_exc.InvalidInput):
            self._test_create_security_rule(protocol=6, port_range_min=-1,
                                            port_range_max=10)

    def test_create_security_rule_egress(self):
        self._test_create_security_rule(protocol=6, direction="egress")

    def test_create_security_rule_max_over_port_max(self):
        with self.assertRaises(n_exc.InvalidInput):
            self._test_create_security_rule(protocol=6, port_range_min=0,
                                            port_range_max=65537)

    def test_create_security_rule_remote_conflicts(self):
        with self.assertRaises(Exception):  # noqa
            self._test_create_security_rule(remote_ip_prefix='192.168.0.1',
                                            remote_group_id='0')

    def test_create_security_rule_no_group(self):
        with self.assertRaises(sg_ext.SecurityGroupNotFound):
            self._test_create_security_rule(group=None)

    def test_create_security_rule_group_at_max(self):
        with self.assertRaises(n_exc.OverQuota):
            self._test_create_security_rule(
                group={'id': 1, 'rules': [models.SecurityGroupRule()]},
                limit_raise=True)

    def test_create_security_group_no_proto_with_ranges_fails(self):
        with self.assertRaises(sg_ext.SecurityGroupProtocolRequiredWithPorts):
            self._test_create_security_rule(protocol=None, port_range_min=0)
        with self.assertRaises(Exception):  # noqa
            self._test_create_security_rule(
                protocol=6, port_range_min=1, port_range_max=0)


class TestQuarkDeleteSecurityGroupRule(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, rule={}, group={'id': 1}):
        dbrule = None
        dbgroup = None
        if group:
            dbgroup = models.SecurityGroup()
            dbgroup.update(group)
        if rule:
            dbrule = models.SecurityGroupRule()
            dbrule.update(dict(rule, group=dbgroup))

        with contextlib.nested(
                mock.patch("quark.db.api.security_group_find"),
                mock.patch("quark.db.api.security_group_rule_find"),
                mock.patch("quark.db.api.security_group_rule_delete"),
        ) as (group_find, rule_find, db_group_delete):
            group_find.return_value = dbgroup
            rule_find.return_value = dbrule
            yield db_group_delete

    def test_delete_security_group_rule(self):
        rule = {'id': 1, 'security_group_id': 1, 'ethertype': 'IPv4',
                'protocol': 6, 'port_range_min': 0, 'port_range_max': 10,
                'direction': 'ingress', 'tenant_id': self.context.tenant_id}

        with self._stubs(dict(rule, group_id=1)) as (db_delete):
            self.plugin.delete_security_group_rule(self.context, 1)
            self.assertTrue(db_delete.called)

    def test_delete_security_group_rule_rule_not_found(self):
        with self._stubs():
            with self.assertRaises(sg_ext.SecurityGroupRuleNotFound):
                self.plugin.delete_security_group_rule(self.context, 1)

    def test_delete_security_group_rule_group_not_found(self):
        rule = {'id': 1, 'security_group_id': 1, 'ethertype': 'IPv4'}
        with self._stubs(dict(rule, group_id=1),
                         None):
            with self.assertRaises(sg_ext.SecurityGroupNotFound):
                self.plugin.delete_security_group_rule(self.context, 1)


class TestQuarkProtocolHandling(test_quark_plugin.TestQuarkPlugin):
    def test_create_security_rule_min_greater_than_max_fails(self):
        with self.assertRaises(sg_ext.SecurityGroupInvalidPortRange):
            protocols.validate_protocol_with_port_ranges(
                ethertype=0x800, protocol=6, port_range_min=10,
                port_range_max=9)

    def test_translate_protocol_string(self):
        proto = protocols.translate_protocol("udp", "IPv4")
        self.assertEqual(proto, 17)

    def test_translate_protocol_int(self):
        proto = protocols.translate_protocol(17, "IPv4")
        self.assertEqual(proto, 17)

    def test_translate_protocol_icmpv4(self):
        proto = protocols.translate_protocol("icmp", "IPv4")
        self.assertEqual(proto, 1)

    def test_translate_protocol_icmpv6(self):
        proto = protocols.translate_protocol("icmp", "IPv6")
        self.assertEqual(proto, 58)

    def test_human_readable_protocol_string(self):
        proto = protocols.human_readable_protocol("UDP", "IPv4")
        self.assertEqual(proto, "UDP")

    def test_human_readable_protocol_int(self):
        proto = protocols.human_readable_protocol(17, "IPv4")
        self.assertEqual(proto, "UDP")

    def test_human_readable_protocol_icmpv4(self):
        proto = protocols.human_readable_protocol(1, "IPv4")
        self.assertEqual(proto, "ICMP")

    def test_human_readable_protocol_icmpv6(self):
        proto = protocols.human_readable_protocol(58, "IPv6")
        self.assertEqual(proto, "ICMP")

    def test_human_readable_protocol_string_as_int(self):
        proto = protocols.human_readable_protocol("17", "IPv4")
        self.assertEqual(proto, "UDP")

    def test_invalid_protocol_string_fail(self):
        with self.assertRaises(sg_ext.SecurityGroupRuleInvalidProtocol):
            protocols.translate_protocol("DERP", "IPv4")

    def test_translate_protocol_under_range(self):
        with self.assertRaises(sg_ext.SecurityGroupRuleInvalidProtocol):
            protocols.translate_protocol(-1, "IPv4")

    def test_translate_protocol_over_range(self):
        with self.assertRaises(sg_ext.SecurityGroupRuleInvalidProtocol):
            protocols.translate_protocol(256, "IPv4")

    def test_translate_protocol_invalid_ethertype(self):
        with self.assertRaises(q_exc.InvalidEthertype):
            protocols.translate_protocol(256, "IPv7")

    def test_validate_remote_ip_prefix(self):
        try:
            protocols.validate_remote_ip_prefix(0x800, "192.168.0.0/24")
        except Exception:
            self.fail("Should not have raised")

    def test_validate_remote_ip_prefix_no_prefix_does_nothing(self):
        try:
            protocols.validate_remote_ip_prefix(0x800, None)
        except Exception:
            self.fail("Should not have raised")

    def test_validate_remote_ip_prefix_ethertype_remote_net_conflict(self):
        with self.assertRaises(n_exc.InvalidInput):
            protocols.validate_remote_ip_prefix(0x86DD, "192.168.0.0/24")


class TestSecurityGroupExceptions(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, finds_rule=True):
        with contextlib.nested(
            mock.patch("quark.db.api.security_group_find"),
            mock.patch("quark.db.api.security_group_rule_find"),
            mock.patch("neutron.extensions.securitygroup."
                       "SecurityGroupNotFound.__init__"),
            mock.patch("neutron.extensions.securitygroup."
                       "SecurityGroupRuleNotFound.__init__"),
            mock.patch("quark.plugin_modules.security_groups."
                       "_validate_security_group_rule")
        ) as (group_find, rule_find, group_exc, rule_exc, validate):
            group_find.return_value = None
            rule_find.return_value = None
            if finds_rule:
                rule_find.return_value = models.SecurityGroupRule()
            group_exc.return_value = None
            rule_exc.return_value = None
            yield group_exc, rule_exc

    def test_delete_group_not_found(self):
        rule = {'id': 1, 'security_group_id': 1, 'ethertype': 'IPv4'}
        with self._stubs() as (group_exc, rule_exc):
            with self.assertRaises(sg_ext.SecurityGroupNotFound):
                self.plugin.delete_security_group(self.context, 1)
                group_exc.assertCalledWith(id=rule["security_group_id"])

    def test_delete_rule_group_not_found(self):
        rule = {'id': 1, 'security_group_id': 1, 'ethertype': 'IPv4'}
        with self._stubs() as (group_exc, rule_exc):
            with self.assertRaises(sg_ext.SecurityGroupNotFound):
                self.plugin.delete_security_group_rule(self.context, 1)
                group_exc.assertCalledWith(id=rule["security_group_id"])

    def test_delete_rule_and_rule_not_found(self):
        rule = {'id': 1, 'security_group_id': 1, 'ethertype': 'IPv4'}
        with self._stubs(False) as (group_exc, rule_exc):
            with self.assertRaises(sg_ext.SecurityGroupRuleNotFound):
                self.plugin.delete_security_group_rule(self.context, 1)
                rule_exc.assertCalledWith(id=rule["id"])

    def test_get_rule_not_found(self):
        rule = {'id': 1, 'security_group_id': 1, 'ethertype': 'IPv4'}
        with self._stubs(False) as (group_exc, rule_exc):
            with self.assertRaises(sg_ext.SecurityGroupRuleNotFound):
                self.plugin.get_security_group_rule(self.context, 1)
                rule_exc.assertCalledWith(id=rule["id"])

    def test_get_group_not_found(self):
        rule = {'id': 1, 'security_group_id': 1, 'ethertype': 'IPv4'}
        with self._stubs() as (group_exc, rule_exc):
            with self.assertRaises(sg_ext.SecurityGroupNotFound):
                self.plugin.get_security_group(self.context, 1)
                rule_exc.assertCalledWith(id=rule["security_group_id"])

    def test_create_rule_group_not_found(self):
        rule = {'id': 1, 'security_group_id': 1, 'ethertype': 'IPv4'}
        with self._stubs() as (group_exc, rule_exc):
            with self.assertRaises(sg_ext.SecurityGroupNotFound):
                self.plugin.create_security_group_rule(
                    self.context, {"security_group_rule": rule})
                rule_exc.assertCalledWith(id=rule["security_group_id"])



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

import netaddr
# NOTE(asadoughi): noqa import lines fix neutron DB changes breaking tests
from neutron.db import agentschedulers_db  # noqa
import neutron.db.model_base
from neutron.db import models_v2 as models
from neutron.db.qos import models as qos_models  # noqa
from neutron.db import rbac_db_models  # noqa
from neutron.db import segments_db  # noqa
from oslo_log import log as logging
from oslo_utils import timeutils
import sqlalchemy as sa
from sqlalchemy.ext import associationproxy
from sqlalchemy.ext import declarative
from sqlalchemy.ext import hybrid
from sqlalchemy import orm

from quark.db import custom_types
from quark.db import ip_types
# NOTE(mdietz): This is the only way to actually create the quotas table,
#              regardless if we need it. This is how it's done upstream.
# NOTE(jhammond): If it isn't obvious quota_driver is unused and that's ok.
#                 DO NOT DELETE IT!!!
from quark import quota_driver  # noqa

HasId = models.HasId

LOG = logging.getLogger(__name__)
TABLE_KWARGS = {"mysql_engine": "InnoDB"}


def _default_list_getset(collection_class, proxy):
    attr = proxy.value_attr

    def getter(obj):
        if obj:
            return getattr(obj, attr, None)
        return []

    if collection_class is dict:
        setter = lambda o, k, v: setattr(o, attr, v)
    else:
        setter = lambda o, v: setattr(o, attr, v)
    return getter, setter


class QuarkBase(neutron.db.model_base.NeutronBaseV2):
    created_at = sa.Column(sa.DateTime(), default=timeutils.utcnow)
    __table_args__ = TABLE_KWARGS


BASEV2 = declarative.declarative_base(cls=QuarkBase)


class TagAssociation(BASEV2, models.HasId):
    __tablename__ = "quark_tag_associations"

    discriminator = sa.Column(sa.String(255))
    tags = associationproxy.association_proxy("tags_association", "tag",
                                              creator=lambda t: Tag(tag=t))

    @classmethod
    def creator(cls, discriminator):
        return lambda tags: TagAssociation(tags=tags,
                                           discriminator=discriminator)

    @property
    def parent(self):
        """Return the parent object."""
        return getattr(self, "%s_parent" % self.discriminator)


class Tag(BASEV2, models.HasId, models.HasTenant):
    __tablename__ = "quark_tags"
    association_uuid = sa.Column(sa.String(36),
                                 sa.ForeignKey(TagAssociation.id),
                                 nullable=False)

    tag = sa.Column(sa.String(255), nullable=False)
    parent = associationproxy.association_proxy("association", "parent")
    association = orm.relationship("TagAssociation",
                                   backref=orm.backref("tags_association"))


class IsHazTags(object):
    @declarative.declared_attr
    def tag_association_uuid(cls):
        return sa.Column(sa.String(36), sa.ForeignKey(TagAssociation.id),
                         nullable=True)

    @declarative.declared_attr
    def tag_association(cls):
        discriminator = cls.__name__.lower()
        creator = TagAssociation.creator(discriminator)
        kwargs = {'creator': creator,
                  'getset_factory': _default_list_getset}
        cls.tags = associationproxy.association_proxy("tag_association",
                                                      "tags", **kwargs)
        backref = orm.backref("%s_parent" % discriminator, uselist=False)
        return orm.relationship("TagAssociation", backref=backref)


class PortIpAssociation(object):
    pass


port_ip_association_table = sa.Table(
    "quark_port_ip_address_associations",
    BASEV2.metadata,
    # We just need sqlalchemy to think these are primary keys
    sa.Column("port_id", sa.String(36),
              sa.ForeignKey("quark_ports.id"), nullable=False,
              primary_key=True),
    sa.Column("ip_address_id", sa.String(36),
              sa.ForeignKey("quark_ip_addresses.id"), nullable=False,
              primary_key=True),
    sa.Column("enabled", sa.Boolean(), default=True, nullable=False,
              server_default='1'),
    sa.Column("service", sa.String(255), default='none', nullable=True,
              server_default='none'),
    **TABLE_KWARGS)


orm.mapper(PortIpAssociation, port_ip_association_table)


class IPAddress(BASEV2, models.HasId):
    """More closely emulate the melange version of the IP table.

    We always mark the record as deallocated rather than deleting it.
    Gives us an IP address owner audit log for free, essentially.
    """
    __tablename__ = "quark_ip_addresses"
    __table_args__ = (sa.UniqueConstraint("subnet_id", "address",
                                          name="subnet_id_address"),
                      TABLE_KWARGS)
    address_readable = sa.Column(sa.String(128), nullable=False)
    address = sa.Column(custom_types.INET(), nullable=False, index=True)
    subnet_id = sa.Column(sa.String(36),
                          sa.ForeignKey("quark_subnets.id",
                                        ondelete="CASCADE"))
    network_id = sa.Column(sa.String(36),
                           sa.ForeignKey("quark_networks.id",
                                         ondelete="CASCADE"))
    network = orm.relationship("Network")
    version = sa.Column(sa.Integer(), index=True)
    allocated_at = sa.Column(sa.DateTime())
    subnet = orm.relationship("Subnet")
    # Need a constant to facilitate the indexed search for new IPs
    _deallocated = sa.Column(sa.Boolean())
    # Legacy data
    used_by_tenant_id = sa.Column(sa.String(255))

    address_type = sa.Column(sa.Enum(ip_types.FIXED, ip_types.FLOATING,
                                     ip_types.SHARED, ip_types.SCALING,
                             name="quark_ip_address_types"))
    associations = orm.relationship(PortIpAssociation, backref="ip_address")
    transaction_id = sa.Column(sa.Integer(),
                               sa.ForeignKey("quark_transactions.id"),
                               nullable=True)
    lock_id = sa.Column(sa.Integer(),
                        sa.ForeignKey("quark_locks.id"),
                        nullable=True)

    def is_shared(self):
        return self.address_type == ip_types.SHARED

    def has_any_shared_owner(self):
        for assoc in self["associations"]:
            if assoc.service != 'none' and assoc.service is not None:
                return True
        return False

    def set_service_for_port(self, port, service):
        for assoc in self["associations"]:
            if assoc.port_id == port["id"]:
                assoc.service = service

    def get_service_for_port(self, port):
        for assoc in self["associations"]:
            if assoc.port_id == port["id"]:
                return assoc.service

    def enabled_for_port(self, port):
        for assoc in self["associations"]:
            if assoc.port_id == port["id"]:
                return assoc.enabled

    @hybrid.hybrid_property
    def deallocated(self):
        return self._deallocated and not self.ports

    @deallocated.setter
    def deallocated(self, val):
        self._deallocated = val
        self.deallocated_at = None
        if val:
            self.deallocated_at = timeutils.utcnow()
            self.allocated_at = None

    # TODO(jkoelker) update the expression to use the jointable as well
    @deallocated.expression
    def deallocated(cls):
        return IPAddress._deallocated

    def formatted(self):
        ip = netaddr.IPAddress(self.address_readable)
        if self.version == 4:
            return str(ip.ipv4())
        return str(ip.ipv6())

    deallocated_at = sa.Column(sa.DateTime(), index=True)
    fixed_ip = None


class FloatingToFixedIPAssociation(object):
    pass

flip_to_fixed_ip_assoc_tbl = sa.Table(
    "quark_floating_to_fixed_ip_address_associations",
    BASEV2.metadata,
    sa.Column("floating_ip_address_id", sa.String(36),
              sa.ForeignKey("quark_ip_addresses.id"), nullable=False,
              primary_key=True),
    sa.Column("fixed_ip_address_id", sa.String(36),
              sa.ForeignKey("quark_ip_addresses.id"), nullable=False,
              primary_key=True),
    sa.Column("enabled", sa.Boolean(), default=True, nullable=False,
              server_default='1'),
    **TABLE_KWARGS)

orm.mapper(FloatingToFixedIPAssociation, flip_to_fixed_ip_assoc_tbl)

IPAddress.fixed_ips = orm.relationship(
    "IPAddress", secondary=flip_to_fixed_ip_assoc_tbl,
    primaryjoin=(IPAddress.id == flip_to_fixed_ip_assoc_tbl
                 .c.floating_ip_address_id and flip_to_fixed_ip_assoc_tbl
                 .c.floating_ip_address_id == 1),
    secondaryjoin=(IPAddress.id == flip_to_fixed_ip_assoc_tbl
                   .c.fixed_ip_address_id), uselist=True)


class Route(BASEV2, models.HasTenant, models.HasId, IsHazTags):
    __tablename__ = "quark_routes"
    cidr = sa.Column(sa.String(64))
    gateway = sa.Column(sa.String(64))
    subnet_id = sa.Column(sa.String(36), sa.ForeignKey("quark_subnets.id",
                                                       ondelete="CASCADE"))


class DNSNameserver(BASEV2, models.HasTenant, models.HasId, IsHazTags):
    __tablename__ = "quark_dns_nameservers"
    ip = sa.Column(custom_types.INET())
    subnet_id = sa.Column(sa.String(36), sa.ForeignKey("quark_subnets.id",
                                                       ondelete="CASCADE"))


def _pools_from_cidr(cidr):
    cidrs = cidr.iter_cidrs()
    if len(cidrs) == 0:
        return []
    if len(cidrs) == 1:
        return [dict(start=str(cidrs[0][0]),
                     end=str(cidrs[0][-1]))]

    pool_start = cidrs[0][0]
    prev_cidr_end = cidrs[0][-1]
    pools = []
    for cidr in cidrs[1:]:
        cidr_start = cidr[0]
        if prev_cidr_end + 1 != cidr_start:
            pools.append(dict(start=str(pool_start),
                              end=str(prev_cidr_end)))
            pool_start = cidr_start
        prev_cidr_end = cidr[-1]
    pools.append(dict(start=str(pool_start), end=str(prev_cidr_end)))
    return pools


class Subnet(BASEV2, models.HasId, IsHazTags):
    """Upstream model for IPs.

    Subnet -> has_many(IPAllocationPool)
    IPAllocationPool -> has_many(IPAvailabilityRange)
        As well as first and last _ip markers for some unknown reason
        first_ip is min(ranges), last_ip is max(ranges)
    IPAvailabilityRange -> belongs_to(IPAllocationPool)
        Also has first and last _ip, but only for the range
    IPAllocation -> belongs_to(port, subnet, network) but NOT IPAllocationPool

    IPAllocationPool and Range seem superfluous. Just create intelligent CIDRs
    for your subnet
    """
    __tablename__ = "quark_subnets"
    id = sa.Column(sa.String(36), primary_key=True)
    name = sa.Column(sa.String(255))
    network_id = sa.Column(sa.String(36), sa.ForeignKey('quark_networks.id'))
    _cidr = sa.Column(sa.String(64), nullable=False)
    _allocation_pool_cache = orm.deferred(sa.Column(sa.Text(), nullable=True))
    tenant_id = sa.Column(sa.String(255), index=True)
    segment_id = sa.Column(sa.String(255), index=True)

    @hybrid.hybrid_property
    def cidr(self):
        return self._cidr

    @hybrid.hybrid_property
    def allocation_pools(self):
        _cache = self.get("_allocation_pool_cache")
        if _cache:
            pools = json.loads(_cache)
            return pools
        else:
            if self["ip_policy"]:
                ip_policy_cidrs = self["ip_policy"].get_cidrs_ip_set()
            else:
                ip_policy_cidrs = netaddr.IPSet([])

            cidr = netaddr.IPSet([netaddr.IPNetwork(self["cidr"])])
            allocatable = cidr - ip_policy_cidrs
            pools = _pools_from_cidr(allocatable)
            return pools

    @cidr.setter
    def cidr(self, val):
        self._cidr = val
        preip = netaddr.IPNetwork(val)
        self.ip_version = preip.version
        ip = netaddr.IPNetwork(val).ipv6()
        self.first_ip = ip.first
        self.last_ip = ip.last
        self.next_auto_assign_ip = self.first_ip

    @cidr.expression
    def cidr(cls):
        return Subnet._cidr

    first_ip = sa.Column(custom_types.INET())
    last_ip = sa.Column(custom_types.INET())
    ip_version = sa.Column(sa.Integer())
    next_auto_assign_ip = sa.Column(custom_types.INET())

    allocated_ips = orm.relationship(IPAddress,
                                     primaryjoin='and_(Subnet.id=='
                                     'IPAddress.subnet_id,'
                                     'IPAddress._deallocated != 1)')
    generated_ips = orm.relationship(IPAddress,
                                     primaryjoin='Subnet.id=='
                                     'IPAddress.subnet_id')
    routes = orm.relationship(Route, primaryjoin="Route.subnet_id==Subnet.id",
                              backref='subnet', cascade='delete')
    enable_dhcp = sa.Column(sa.Boolean(), default=False)
    dns_nameservers = orm.relationship(
        DNSNameserver,
        primaryjoin="DNSNameserver.subnet_id==Subnet.id",
        backref='subnet',
        cascade='delete')
    ip_policy_id = sa.Column(sa.String(36),
                             sa.ForeignKey("quark_ip_policy.id"))
    # Legacy data
    do_not_use = sa.Column(sa.Boolean(), default=False)


port_group_association_table = sa.Table(
    "quark_port_security_group_associations",
    BASEV2.metadata,
    sa.Column("port_id", sa.String(36),
              sa.ForeignKey("quark_ports.id")),
    sa.Column("group_id", sa.String(36),
              sa.ForeignKey("quark_security_groups.id")),
    **TABLE_KWARGS)


class SecurityGroupRule(BASEV2, models.HasId, models.HasTenant):
    __tablename__ = "quark_security_group_rules"
    id = sa.Column(sa.String(36), primary_key=True)
    group_id = sa.Column(sa.String(36),
                         sa.ForeignKey("quark_security_groups.id"),
                         nullable=False)
    direction = sa.Column(sa.String(10), nullable=False)
    ethertype = sa.Column(sa.Integer(), nullable=False)
    port_range_max = sa.Column(sa.Integer(), nullable=True)
    port_range_min = sa.Column(sa.Integer(), nullable=True)
    protocol = sa.Column(sa.Integer(), nullable=True)
    remote_ip_prefix = sa.Column(sa.String(255), nullable=True)
    remote_group_id = sa.Column(sa.String(36),
                                sa.ForeignKey("quark_security_groups.id"),
                                nullable=True)


class SecurityGroup(BASEV2, models.HasId):
    __tablename__ = "quark_security_groups"
    id = sa.Column(sa.String(36), primary_key=True)
    name = sa.Column(sa.String(255), nullable=False)
    description = sa.Column(sa.String(255), nullable=False)
    join = "SecurityGroupRule.group_id==SecurityGroup.id"
    rules = orm.relationship(SecurityGroupRule, backref='group',
                             cascade='delete',
                             primaryjoin=join)
    tenant_id = sa.Column(sa.String(255), index=True)


class Port(BASEV2, models.HasTenant, models.HasId, IsHazTags):
    __tablename__ = "quark_ports"
    id = sa.Column(sa.String(36), primary_key=True)
    name = sa.Column(sa.String(255), index=True)
    admin_state_up = sa.Column(sa.Boolean(), default=True)
    network_id = sa.Column(sa.String(36), sa.ForeignKey("quark_networks.id"),
                           nullable=False)

    backend_key = sa.Column(sa.String(36), nullable=False)
    mac_address = sa.Column(sa.BigInteger())
    device_id = sa.Column(sa.String(255), nullable=False, index=True)
    device_owner = sa.Column(sa.String(255))
    bridge = sa.Column(sa.String(255))
    associations = orm.relationship(PortIpAssociation, backref="port")

    network_plugin = sa.Column(sa.String(36), nullable=True)

    @declarative.declared_attr
    def ip_addresses(cls):
        primaryjoin = cls.id == port_ip_association_table.c.port_id
        secondaryjoin = (port_ip_association_table.c.ip_address_id ==
                         IPAddress.id)
        return orm.relationship(IPAddress, primaryjoin=primaryjoin,
                                secondaryjoin=secondaryjoin,
                                secondary=port_ip_association_table,
                                backref='ports',
                                order_by='IPAddress.allocated_at')

    @declarative.declared_attr
    def security_groups(cls):
        primaryjoin = cls.id == port_group_association_table.c.port_id
        secondaryjoin = (port_group_association_table.c.group_id ==
                         SecurityGroup.id)
        return orm.relationship(SecurityGroup, primaryjoin=primaryjoin,
                                secondaryjoin=secondaryjoin,
                                secondary=port_group_association_table,
                                backref="ports")


# Indices tailored specifically to get_instance_nw_info calls from nova
sa.Index("idx_ports_1", Port.__table__.c.device_id, Port.__table__.c.tenant_id)
sa.Index("idx_ports_2", Port.__table__.c.device_owner,
         Port.__table__.c.network_id)
sa.Index("idx_ports_3", Port.__table__.c.tenant_id)


class MacAddress(BASEV2, models.HasTenant):
    __tablename__ = "quark_mac_addresses"
    address = sa.Column(sa.BigInteger(), primary_key=True)
    mac_address_range_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("quark_mac_address_ranges.id", ondelete="CASCADE"),
        nullable=False)
    deallocated = sa.Column(sa.Boolean(), index=True)
    deallocated_at = sa.Column(sa.DateTime(), index=True)
    orm.relationship(Port, backref="mac_address")
    transaction_id = sa.Column(sa.Integer(),
                               sa.ForeignKey("quark_transactions.id"),
                               nullable=True)


class MacAddressRange(BASEV2, models.HasId):
    __tablename__ = "quark_mac_address_ranges"
    cidr = sa.Column(sa.String(255), nullable=False)
    first_address = sa.Column(sa.BigInteger(), nullable=False)
    last_address = sa.Column(sa.BigInteger(), nullable=False)
    next_auto_assign_mac = sa.Column(sa.BigInteger(), nullable=False)
    allocated_macs = orm.relationship(MacAddress,
                                      primaryjoin='and_(MacAddressRange.id=='
                                      'MacAddress.mac_address_range_id, '
                                      'MacAddress.deallocated!=1)',
                                      backref="mac_address_range")
    do_not_use = sa.Column(sa.Boolean(), default=False, nullable=False,
                           server_default='0')


class IPPolicy(BASEV2, models.HasId, models.HasTenant):
    __tablename__ = "quark_ip_policy"
    networks = orm.relationship(
        "Network",
        primaryjoin="IPPolicy.id==Network.ip_policy_id",
        backref="ip_policy")
    subnets = orm.relationship(
        "Subnet",
        primaryjoin="IPPolicy.id==Subnet.ip_policy_id",
        backref="ip_policy")
    exclude = orm.relationship(
        "IPPolicyCIDR",
        primaryjoin="IPPolicy.id==IPPolicyCIDR.ip_policy_id",
        backref="ip_policy")
    name = sa.Column(sa.String(255), nullable=True)
    description = sa.Column(sa.String(255), nullable=True)
    size = sa.Column(custom_types.INET())

    def get_cidrs_ip_set(self):
        ip_policies = self.get("exclude", [])
        ip_policy_cidrs = [ip_policy.cidr for ip_policy in ip_policies]
        return netaddr.IPSet(ip_policy_cidrs)


class IPPolicyCIDR(BASEV2, models.HasId):
    __tablename__ = "quark_ip_policy_cidrs"
    ip_policy_id = sa.Column(sa.String(36), sa.ForeignKey(
        "quark_ip_policy.id", ondelete="CASCADE"))
    cidr = sa.Column(sa.String(64))
    first_ip = sa.Column(custom_types.INET())
    last_ip = sa.Column(custom_types.INET())


class Network(BASEV2, models.HasId):
    __tablename__ = "quark_networks"
    name = sa.Column(sa.String(255))
    ports = orm.relationship(Port, backref='network')
    subnets = orm.relationship(Subnet, backref='network')
    ip_policy_id = sa.Column(sa.String(36),
                             sa.ForeignKey("quark_ip_policy.id"))
    network_plugin = sa.Column(sa.String(36))
    ipam_strategy = sa.Column(sa.String(255))
    tenant_id = sa.Column(sa.String(255), index=True)


class Transaction(BASEV2):
    __tablename__ = "quark_transactions"
    id = sa.Column(sa.Integer, primary_key=True)


class Lock(BASEV2):
    __tablename__ = "quark_locks"
    id = sa.Column(sa.Integer, primary_key=True)
    type = sa.Column(sa.Enum("ip_address"), nullable=False)


class LockHolder(BASEV2):
    __tablename__ = "quark_lock_holders"
    id = sa.Column(sa.Integer, primary_key=True)
    lock_id = sa.Column(sa.Integer(),
                        sa.ForeignKey("quark_locks.id"),
                        nullable=False)
    name = sa.Column(sa.String(255), nullable=True)


class SegmentAllocation(BASEV2):
    """A segment allocation."""
    __tablename__ = "quark_segment_allocations"

    # a particular segment id is unique across the segment and type - this data
    # is denormalized to give us some safety around allocations, as well
    # as allow us to look up allocations to reallocate without a join.
    id = sa.Column(sa.BigInteger(), primary_key=True, autoincrement=False)
    segment_id = sa.Column(sa.String(36), primary_key=True)
    segment_type = sa.Column(sa.String(36), primary_key=True)

    segment_allocation_range_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("quark_segment_allocation_ranges.id",
                      ondelete="CASCADE"))

    network_id = sa.Column(sa.String(36), nullable=True)

    deallocated = sa.Column(sa.Boolean(), index=True)
    deallocated_at = sa.Column(sa.DateTime(), index=True)


class SegmentAllocationRange(BASEV2, models.HasId):
    """Ranges of space for segment ids available for allocation."""
    __tablename__ = "quark_segment_allocation_ranges"
    segment_id = sa.Column(sa.String(36), index=True)
    segment_type = sa.Column(sa.String(36), index=True)

    first_id = sa.Column(sa.BigInteger(), nullable=False)
    last_id = sa.Column(sa.BigInteger(), nullable=False)

    do_not_use = sa.Column(sa.Boolean(), default=False, nullable=False)


class AsyncTransactions(BASEV2, models.HasId):
    __tablename__ = "quark_async_transactions"
    tenant_id = sa.Column(sa.String(255), index=True)
    action = sa.Column(sa.String(255))
    completed = sa.Column(sa.Boolean(), default=False)



import mock
import netaddr

import contextlib

from quark.db import api as db_api
from quark import exceptions
import quark.ipam
import quark.plugin
import quark.plugin_modules.ip_addresses as ip_api
import quark.plugin_modules.mac_address_ranges as macrng_api
import quark.plugin_modules.networks as network_api
import quark.plugin_modules.ports as port_api
import quark.plugin_modules.subnets as subnet_api
from quark.tests.functional.mysql.base import MySqlBaseFunctionalTest


class QuarkUpdatePorts(MySqlBaseFunctionalTest):
    @contextlib.contextmanager
    def _stubs(self, network_info, subnet_info, port_info):
        self.ipam = quark.ipam.QuarkIpamANY()
        with contextlib.nested(
                mock.patch("neutron.common.rpc.get_notifier"),
                mock.patch("neutron.quota.QUOTAS.limit_check")):
            net = network_api.create_network(self.context, network_info)
            mac = {'mac_address_range': dict(cidr="AA:BB:CC")}
            self.context.is_admin = True
            macrng_api.create_mac_address_range(self.context, mac)
            self.context.is_admin = False
            subnet_info['subnet']['network_id'] = net['id']
            port_info['port']['network_id'] = net['id']
            sub = subnet_api.create_subnet(self.context, subnet_info)
            port = port_api.create_port(self.context, port_info)
            yield net, sub, port

    def test_update_fixed_ips_regression_RM9097(self):
        cidr = "192.168.1.0/24"
        ip_network = netaddr.IPNetwork(cidr)
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        network = {"network": network}
        subnet = dict(id=1, ip_version=4, next_auto_assign_ip=2,
                      cidr=cidr, first_ip=ip_network.first,
                      last_ip=ip_network.last, ip_policy=None,
                      tenant_id="fake")
        subnet = {"subnet": subnet}
        port = {"port": dict()}

        def _make_body(ip):
            fix_ip = dict(ip_address=ip, subnet_id=sub['id'])
            port_info = {"port": dict(fixed_ips=[fix_ip])}
            return port_info

        with self._stubs(network, subnet, port) as (net, sub, port):
            id = port['id']

            ip = "192.168.1.50"
            port = port_api.update_port(self.context, id, _make_body(ip))
            self.assertEqual(ip, port['fixed_ips'][0]['ip_address'])

            with self.assertRaises(exceptions.IPAddressNotInSubnet):
                ip = "192.168.2.50"
                port = port_api.update_port(self.context, id, _make_body(ip))
                self.assertEqual(ip, port['fixed_ips'][0]['ip_address'])

            ip = "192.168.1.75"
            port = port_api.update_port(self.context, id, _make_body(ip))
            self.assertEqual(ip, port['fixed_ips'][0]['ip_address'])

            ip = "192.168.1.50"
            port = port_api.update_port(self.context, id, _make_body(ip))
            self.assertEqual(ip, port['fixed_ips'][0]['ip_address'])


class QuarkFindPorts(MySqlBaseFunctionalTest):
    def test_ip_address_port_find_service(self):
        net = db_api.network_create(self.context)
        port = db_api.port_create(self.context, network_id=net["id"],
                                  backend_key="", device_id="")
        ip_address = db_api.ip_address_create(
            self.context, address=netaddr.IPAddress("0.0.0.0"))
        self.context.session.flush()

        ip_address = db_api.port_associate_ip(self.context, [port], ip_address)
        ip_address.set_service_for_port(port, "foobar")
        self.context.session.flush()

        ports = ip_api.get_ports_for_ip_address(
            self.context, ip_address["id"],
            filters={"service": "not-foobar"})
        self.assertEqual(len(ports), 0)

        ports = ip_api.get_ports_for_ip_address(
            self.context, ip_address["id"],
            filters={"service": "foobar"})
        self.assertEqual(len(ports), 1)



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#  under the License.

import contextlib

import mock
from neutron.common import exceptions as n_exc_ext
from neutron_lib import exceptions as n_exc

from quark import exceptions as q_exc
from quark.plugin_modules import ip_policies as ippol
from quark.tests import test_base
from quark.tests import test_quark_plugin


class TestQuarkGetIpPolicies(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, ip_policy):
        db_mod = "quark.db.api"
        with mock.patch("%s.ip_policy_find" % db_mod) as ip_policy_find:
            ip_policy_find.return_value = ip_policy
            yield

    def test_get_ip_policy_not_found(self):
        with self._stubs(None):
            with self.assertRaises(q_exc.IPPolicyNotFound):
                self.plugin.get_ip_policy(self.context, 1)

    def test_get_ip_policy(self):
        ip_policy = dict(
            id=1,
            tenant_id=1,
            name="foo",
            subnets=[dict(id=1)],
            networks=[dict(id=2)],
            exclude=[dict(cidr="0.0.0.0/32")])
        with self._stubs(ip_policy):
            resp = self.plugin.get_ip_policy(self.context, 1)
            self.assertEqual(len(resp.keys()), 6)
            self.assertEqual(resp["id"], 1)
            self.assertEqual(resp["name"], "foo")
            self.assertEqual(resp["subnet_ids"], [1])
            self.assertEqual(resp["network_ids"], [2])
            self.assertEqual(resp["exclude"], ["0.0.0.0/32"])
            self.assertEqual(resp["tenant_id"], 1)

    def test_get_ip_policies(self):
        ip_policy = dict(
            id=1,
            tenant_id=1,
            name="foo",
            subnets=[dict(id=1)],
            networks=[dict(id=2)],
            exclude=[dict(cidr="0.0.0.0/32")])
        with self._stubs([ip_policy]):
            resp = self.plugin.get_ip_policies(self.context)
            self.assertEqual(len(resp), 1)
            resp = resp[0]
            self.assertEqual(len(resp.keys()), 6)
            self.assertEqual(resp["id"], 1)
            self.assertEqual(resp["subnet_ids"], [1])
            self.assertEqual(resp["network_ids"], [2])
            self.assertEqual(resp["exclude"], ["0.0.0.0/32"])
            self.assertEqual(resp["name"], "foo")
            self.assertEqual(resp["tenant_id"], 1)


class TestQuarkCreateIpPolicies(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, ip_policy, subnets=None, nets=None):
        db_mod = "quark.db.api"
        with contextlib.nested(
            mock.patch("%s.subnet_find" % db_mod),
            mock.patch("%s.network_find" % db_mod),
            mock.patch("%s.ip_policy_create" % db_mod),
            mock.patch("%s.route_find" % db_mod)
        ) as (subnet_find, net_find, ip_policy_create, route_find):
            subnet_find.return_value = subnets if subnets else None
            net_find.return_value = nets if nets else None
            ip_policy_create.return_value = ip_policy
            route_find.return_value = [{"nexthop": "1.2.3.4"}]
            yield ip_policy_create

    def test_create_ip_policy_invalid_body_missing_exclude(self):
        with self._stubs(None):
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.create_ip_policy(self.context, dict(
                    ip_policy=dict()))

    def test_create_ip_policy_with_both_network_and_subnet_ids(self):
        with self._stubs(None):
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.create_ip_policy(self.context, dict(
                    ip_policy=dict(network_ids=[1], subnet_ids=[1])))

    def test_create_ip_policy_invalid_body_missing_netsubnet(self):
        with self._stubs(None):
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.create_ip_policy(self.context, dict(
                    ip_policy=dict(exclude=["1.1.1.1/24"])))

    def test_create_ip_policy_invalid_subnet(self):
        with self._stubs(None):
            with self.assertRaises(n_exc.SubnetNotFound):
                self.plugin.create_ip_policy(self.context, dict(
                    ip_policy=dict(subnet_ids=[1],
                                   exclude=["1.1.1.1/24"])))

    def test_create_ip_policy_invalid_network(self):
        with self._stubs(None):
            with self.assertRaises(n_exc.NetworkNotFound):
                self.plugin.create_ip_policy(self.context, dict(
                    ip_policy=dict(network_ids=[1],
                                   exclude=["1.1.1.1/24"])))

    def test_create_ip_policy_network_ip_policy_already_exists(self):
        with self._stubs(None, nets=[dict(id=1, ip_policy=dict(id=2),
                                          subnets=[dict(id=1,
                                                        cidr="1.1.1.1/16")])]):
            with self.assertRaises(q_exc.IPPolicyAlreadyExists):
                self.plugin.create_ip_policy(self.context, dict(
                    ip_policy=dict(network_ids=[1],
                                   exclude=["1.1.1.1/24"])))

    def test_create_ip_policy_subnet_ip_policy_already_exists(self):
        with self._stubs(None, subnets=[dict(id=1, ip_policy=dict(id=2),
                                             cidr="1.1.1.1/16")]):
            with self.assertRaises(q_exc.IPPolicyAlreadyExists):
                self.plugin.create_ip_policy(self.context, dict(
                    ip_policy=dict(subnet_ids=[1],
                                   exclude=["1.1.1.1/24"])))

    def test_create_ip_policy_network(self):
        ipp = dict(subnet_id=None, network_id=1,
                   exclude=["1.1.1.1/24"])
        with self._stubs(ipp, nets=[dict(id=1, ip_policy=dict(id=2),
                                         subnets=[dict(id=1,
                                                       cidr="1.1.1.1/16")])]):
            with self.assertRaises(q_exc.IPPolicyAlreadyExists):
                self.plugin.create_ip_policy(self.context, dict(
                    ip_policy=dict(network_ids=[ipp["network_id"]],
                                   exclude=ipp["exclude"])))

    def test_create_ip_policy_subnet(self):
        ipp = dict(subnet_id=1, network_id=None,
                   exclude=["1.1.1.1/24"])
        with self._stubs(ipp, subnets=[dict(id=1, ip_policy=dict(id=2),
                                            cidr="1.1.1.1/16")]):
            with self.assertRaises(q_exc.IPPolicyAlreadyExists):
                self.plugin.create_ip_policy(self.context, dict(
                    ip_policy=dict(subnet_ids=[ipp["subnet_id"]],
                                   exclude=ipp["exclude"])))

    def test_create_ip_policy_with_cidr_that_does_not_fit_into_subnet(self):
        ipp = dict(
            subnets=[dict(id=1, version=4, cidr="192.168.1.1/24")],
            networks=[],
            id=1,
            tenant_id=1,
            exclude=["10.10.10.100/32"],
            name="foo")
        with self._stubs(ipp,
                         subnets=[dict(id=1, ip_policy=None,
                                       version=ipp["subnets"][0]["version"],
                                       cidr=ipp["subnets"][0]["cidr"])]):
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.create_ip_policy(self.context, dict(
                    ip_policy=dict(subnet_ids=[1],
                                   exclude=ipp["exclude"])))

    def test_create_ip_policy_with_ipv6_subnet_cidr(self):
        ipp = dict(
            subnets=[dict(id=1, version=6, cidr='::/64')],
            networks=[],
            id=1,
            tenant_id=1,
            exclude=[dict(cidr="::/128")],
            name="foo")
        with self._stubs(ipp,
                         subnets=[dict(id=1, ip_policy=None,
                                       version=ipp["subnets"][0]["version"],
                                       cidr=ipp["subnets"][0]["cidr"])]):
            exclude = [ippc["cidr"] for ippc in ipp["exclude"]]
            resp = self.plugin.create_ip_policy(self.context, dict(
                ip_policy=dict(subnet_ids=[1], exclude=exclude)))
            self.assertEqual(len(resp.keys()), 6)
            self.assertEqual(resp["subnet_ids"], [1])
            self.assertEqual(resp["network_ids"], [])
            # NOTE(jmeridth): below is mocked that way, so it won't get
            # additional default policies in exclude
            # ippol.ensure_default_policy is tested below in this file
            self.assertEqual(resp["exclude"], ["::/128"])
            self.assertEqual(resp["name"], "foo")
            self.assertEqual(resp["tenant_id"], 1)

    def test_create_ip_policy(self):
        ipp = dict(
            subnets=[dict(id=1, cidr='0.0.0.0/16')],
            networks=[],
            id=1,
            tenant_id=1,
            exclude=[dict(cidr="0.0.0.0/24")],
            name="foo")
        with self._stubs(ipp, subnets=[dict(
                id=1, ip_policy=None, cidr=ipp["subnets"][0]["cidr"])]):
            exclude = [ippc["cidr"] for ippc in ipp["exclude"]]
            resp = self.plugin.create_ip_policy(self.context, dict(
                ip_policy=dict(subnet_ids=[1], exclude=exclude)))
            self.assertEqual(len(resp.keys()), 6)
            self.assertEqual(resp["subnet_ids"], [1])
            self.assertEqual(resp["network_ids"], [])
            # NOTE(jmeridth): below is mocked that way, so it won't get
            # additional default policies in exclude
            # ippol.ensure_default_policy is tested below in this file
            self.assertEqual(resp["exclude"], ["0.0.0.0/24"])
            self.assertEqual(resp["name"], "foo")
            self.assertEqual(resp["tenant_id"], 1)

    def test_create_ip_policy_only_called_once_with_multiple_networks(self):
        ipp = dict(
            subnets=[],
            networks=[dict(id=1, subnets=[dict(id=1,
                           ip_policy=None, cidr='0.0.0.0/24')]),
                      dict(id=2, subnets=[dict(id=2,
                           ip_policy=None, cidr='0.0.0.0/24')])],
            id=1,
            tenant_id=1,
            exclude=[dict(cidr="0.0.0.1/32")],
            name="foo")
        with self._stubs(ipp, nets=ipp["networks"]) as (ip_policy_create):
            resp = self.plugin.create_ip_policy(self.context, dict(
                ip_policy=dict(network_ids=[1, 2], exclude=["0.0.0.1/32"])))
            exclude = ['0.0.0.1/32', '0.0.0.0/32', '0.0.0.255/32']
            ip_policy_create.assert_called_once_with(
                self.context, exclude=exclude,
                networks=[{'subnets':
                          [{'cidr': '0.0.0.0/24', 'ip_policy': None,
                            'id': 1}], 'id': 1},
                          {'subnets':
                          [{'cidr': '0.0.0.0/24', 'ip_policy': None,
                            'id': 2}], 'id': 2}])
            self.assertEqual(len(resp.keys()), 6)
            self.assertEqual(resp["subnet_ids"], [])
            self.assertEqual(resp["network_ids"], [1, 2])
            # NOTE(jmeridth): below is mocked that way, so it won't get
            # additional default policies in exclude
            # ippol.ensure_default_policy is tested below in this file
            self.assertEqual(resp["exclude"], ["0.0.0.1/32"])
            self.assertEqual(resp["name"], "foo")
            self.assertEqual(resp["tenant_id"], 1)

    def test_create_ip_policy_only_called_once_with_multiple_subnets(self):
        ipp = dict(
            subnets=[dict(id=3, cidr='0.0.0.0/16'),
                     dict(id=4, cidr='0.0.0.0/16')],
            networks=[],
            id=1,
            tenant_id=1,
            exclude=[dict(cidr="0.0.0.1/32")],
            name="foo")
        with self._stubs(ipp, subnets=ipp["subnets"]) as (ip_policy_create):
            resp = self.plugin.create_ip_policy(self.context, dict(
                ip_policy=dict(subnet_ids=[3, 4], exclude=["0.0.0.1/32"])))
            exclude = ['0.0.0.1/32', '0.0.0.0/32', '0.0.255.255/32']
            ip_policy_create.assert_called_once_with(
                self.context, exclude=exclude,
                subnets=[{'cidr': '0.0.0.0/16', 'id': 3},
                         {'cidr': '0.0.0.0/16', 'id': 4}])
            self.assertEqual(len(resp.keys()), 6)
            self.assertEqual(resp["subnet_ids"], [3, 4])
            self.assertEqual(resp["network_ids"], [])
            # NOTE(jmeridth): below is mocked that way, so it won't get
            # additional default policies in exclude
            # ippol.ensure_default_policy is tested below in this file
            self.assertEqual(resp["exclude"], ["0.0.0.1/32"])
            self.assertEqual(resp["name"], "foo")
            self.assertEqual(resp["tenant_id"], 1)


class TestQuarkUpdateIpPolicies(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, ip_policy, subnets=None, networks=None):
        if not subnets:
            subnets = []
        if not networks:
            networks = []
        db_mod = "quark.db.api"
        with contextlib.nested(
            mock.patch("%s.ip_policy_find" % db_mod),
            mock.patch("%s.subnet_find" % db_mod),
            mock.patch("%s.network_find" % db_mod),
            mock.patch("%s.ip_policy_update" % db_mod),
        ) as (ip_policy_find, subnet_find, network_find, ip_policy_update):
            ip_policy_find.return_value = ip_policy
            subnet_find.return_value = subnets
            network_find.return_value = networks
            yield ip_policy_update

    def test_update_ip_policy_not_found(self):
        with self._stubs(None):
            with self.assertRaises(q_exc.IPPolicyNotFound):
                self.plugin.update_ip_policy(self.context, 1,
                                             dict(ip_policy=None))

    def test_update_ip_policy_with_both_network_and_subnet_ids(self):
        ipp = dict(id=1, subnets=[])
        with self._stubs(ipp):
            with self.assertRaises(n_exc.BadRequest):
                self.plugin.update_ip_policy(self.context, 1, dict(
                    ip_policy=dict(network_ids=[1], subnet_ids=[1])))

    def test_update_ip_policy_subnets_not_found(self):
        ipp = dict(id=1, subnets=[])
        with self._stubs(ipp):
            with self.assertRaises(n_exc.SubnetNotFound):
                self.plugin.update_ip_policy(
                    self.context,
                    1,
                    dict(ip_policy=dict(subnet_ids=[100])))

    def test_update_ip_policy_subnets_already_exists(self):
        ipp = dict(id=1, subnets=[dict()])
        with self._stubs(
            ipp, subnets=[dict(id=1, ip_policy=dict(id=1))]
        ):
            with self.assertRaises(q_exc.IPPolicyAlreadyExists):
                self.plugin.update_ip_policy(
                    self.context,
                    1,
                    dict(ip_policy=dict(subnet_ids=[100])))

    def test_update_ip_policy_subnets(self):
        ipp = dict(id=1, subnets=[dict()],
                   exclude=["0.0.0.0/24"],
                   name="foo", tenant_id=1)
        with self._stubs(
            ipp, subnets=[dict(id=1, ip_policy=None)]
        ) as (ip_policy_update):
            self.plugin.update_ip_policy(
                self.context,
                1,
                dict(ip_policy=dict(subnet_ids=[100])))
            self.assertEqual(ip_policy_update.called, 1)

    def test_update_ip_policy_subnets_empty_exclude(self):
        ipp = dict(id=1, subnets=[dict()],
                   exclude=["0.0.0.40/32"],
                   name="foo", tenant_id=1)
        with self._stubs(
            ipp, subnets=[dict(id=1, cidr="0.0.0.0/16", ip_policy=None)]
        ) as (ip_policy_update):
            self.plugin.update_ip_policy(
                self.context,
                1,
                dict(ip_policy=dict(subnet_ids=[100], exclude=[])))
            ip_policy_update.assert_called_once_with(
                self.context, ipp, subnet_ids=[100], exclude=[
                    "0.0.0.0/32", "0.0.255.255/32"])

    def test_update_ip_policy_subnets_empty_exclude_without_subnet_ids(self):
        ipp = dict(id=1, subnets=[dict(cidr="0.0.0.0/16")],
                   exclude=["0.0.0.40/32"],
                   name="foo", tenant_id=1)
        with self._stubs(ipp) as (ip_policy_update):
            self.plugin.update_ip_policy(
                self.context,
                1,
                dict(ip_policy=dict(exclude=[])))
            ip_policy_update.assert_called_once_with(
                self.context, ipp, exclude=["0.0.0.0/32", "0.0.255.255/32"])

    def test_update_ip_policy_networks_not_found(self):
        ipp = dict(id=1, networks=[])
        with self._stubs(ipp):
            with self.assertRaises(n_exc.NetworkNotFound):
                self.plugin.update_ip_policy(
                    self.context,
                    1,
                    dict(ip_policy=dict(network_ids=[100])))

    def test_update_ip_policy_networks(self):
        ipp = dict(id=1, networks=[dict()],
                   exclude=["0.0.0.0/24"],
                   name="foo", tenant_id=1)
        with self._stubs(
            ipp, networks=[dict(id=1, ip_policy=None)]
        ) as (ip_policy_update):
            self.plugin.update_ip_policy(
                self.context,
                1,
                dict(ip_policy=dict(network_ids=[100])))
            self.assertEqual(ip_policy_update.called, 1)

    def test_update_ip_policy_exclude_v4(self):
        subnets = [dict(id=100, cidr="0.0.0.0/16")]
        ipp = dict(id=1, subnets=subnets,
                   exclude=["0.0.0.0/24"],
                   name="foo", tenant_id=1)
        with self._stubs(ipp, subnets=subnets) as (ip_policy_update):
            self.plugin.update_ip_policy(
                self.context,
                1,
                dict(ip_policy=dict(subnet_ids=[100], exclude=["0.0.0.1/32"])))
            ip_policy_update.assert_called_once_with(
                self.context,
                ipp,
                subnet_ids=[100],
                exclude=["0.0.0.1/32", "0.0.0.0/32", "0.0.255.255/32"])

    def test_update_ip_policy_exclude_v6(self):
        subnets = [dict(id=100, cidr="::/64")]
        ipp = dict(id=1, subnets=subnets,
                   exclude=["::/128"],
                   name="foo", tenant_id=1)
        with self._stubs(ipp, subnets=subnets) as (ip_policy_update):
            self.plugin.update_ip_policy(
                self.context,
                1,
                dict(ip_policy=dict(subnet_ids=[100], exclude=["::1/128"])))
            ip_policy_update.assert_called_once_with(
                self.context,
                ipp,
                subnet_ids=[100],
                exclude=["::1/128", "::/128", "::ffff:ffff:ffff:ffff/128"])


class TestQuarkDeleteIpPolicies(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, ip_policy):
        db_mod = "quark.db.api"
        with contextlib.nested(
            mock.patch("%s.ip_policy_find" % db_mod),
            mock.patch("%s.ip_policy_delete" % db_mod),
        ) as (ip_policy_find, ip_policy_delete):
            ip_policy_find.return_value = ip_policy
            yield ip_policy_find, ip_policy_delete

    def test_delete_ip_policy_not_found(self):
        with self._stubs(None):
            with self.assertRaises(q_exc.IPPolicyNotFound):
                self.plugin.delete_ip_policy(self.context, 1)

    def test_delete_ip_policy_in_use(self):
        with self._stubs(dict(networks=True)):
            with self.assertRaises(q_exc.IPPolicyInUse):
                self.plugin.delete_ip_policy(self.context, 1)

    def test_delete_ip_policy(self):
        ip_policy = dict(
            id=1,
            networks=[],
            subnets=[])
        with self._stubs(ip_policy) as (ip_policy_find, ip_policy_delete):
            self.plugin.delete_ip_policy(self.context, 1)
            self.assertEqual(ip_policy_find.call_count, 1)
            self.assertEqual(ip_policy_delete.call_count, 1)


class TestQuarkUpdatePolicySubnetWithRoutes(test_quark_plugin.TestQuarkPlugin):
    @contextlib.contextmanager
    def _stubs(self, ip_policy, subnets=None, routes=None):
        subnets = subnets or []
        db_mod = "quark.db.api"
        with contextlib.nested(
            mock.patch("%s.ip_policy_find" % db_mod),
            mock.patch("%s.subnet_find" % db_mod),
            mock.patch("%s.route_find" % db_mod),
            mock.patch("%s.ip_policy_update" % db_mod),
        ) as (ip_policy_find, subnet_find, route_find, ip_policy_update):
            ip_policy_find.return_value = ip_policy
            subnet_find.return_value = subnets
            route_find.return_value = routes
            yield ip_policy_update

    def test_update_ip_policy_has_route_conflict_raises(self):
        subnet = dict(id=1, cidr="192.168.0.0/24")
        ipp = dict(id=1, subnets=[subnet], exclude=["192.168.0.1/32"],
                   name="foo", tenant_id=1)
        route = {"gateway": "192.168.0.1", "subnet_id": subnet["id"]}
        with self._stubs(ipp, subnets=[subnet], routes=[route]):
            with self.assertRaises(
                    n_exc_ext.GatewayConflictWithAllocationPools):
                self.plugin.update_ip_policy(
                    self.context, 1,
                    dict(ip_policy=dict(subnet_ids=[1], exclude=[])))

    def test_update_ip_policy_no_route_conflict(self):
        subnet = dict(id=1, cidr="192.168.0.0/24")
        ipp = dict(id=1, subnets=[subnet], exclude=["192.168.0.1/32"],
                   name="foo", tenant_id=1)
        route = {"gateway": "192.168.0.1", "subnet_id": subnet["id"]}
        with self._stubs(ipp, subnets=[subnet], routes=[route]):
            try:
                self.plugin.update_ip_policy(
                    self.context, 1,
                    dict(ip_policy=dict(subnet_ids=[1],
                                        exclude=["192.168.0.0/24"])))
            except Exception as e:
                self.fail("This shouldn't have raised: %s" % e)


class TestQuarkValidateCIDRsFitsIntoSubnets(test_quark_plugin.TestQuarkPlugin):
    def test_normal_cidr_and_valid_subnet(self):
        try:
            ippol._validate_cidrs_fit_into_subnets(
                ["192.168.0.100/32"],
                [dict(id=1, cidr="192.168.0.0/24")])
        except Exception:
            self.fail("Should not have failed")

    def test_normal_ipv4_cidr_and_valid_ipv6_subnet(self):
        try:
            ippol._validate_cidrs_fit_into_subnets(
                ["192.168.0.100/32"], [dict(id=1, cidr="::/96")])
        except Exception:
            self.fail("Should not have failed")

    def test_normal_ipv6_cidr_and_valid_ipv6_subnet(self):
        try:
            ippol._validate_cidrs_fit_into_subnets(
                ["::/128"], [dict(id=1, cidr="::/96")])
        except Exception:
            self.fail("Should not have failed")

    def test_normal_ipv6_cidr_and_valid_ipv4_subnet(self):
        try:
            ippol._validate_cidrs_fit_into_subnets(
                ["::/128"], [dict(id=1, cidr="192.168.0.0/24")])
        except Exception:
            self.fail("Should not have failed")

    def test_normal_cidr_and_multiple_valid_subnet(self):
        try:
            ippol._validate_cidrs_fit_into_subnets(
                ["192.168.0.100/32"],
                [dict(id=1, cidr="192.168.0.0/24"),
                 dict(id=2, cidr="192.168.0.0/16")])
        except Exception:
            self.fail("Should not have failed")

    def test_normal_ipv6_cidr_and_multiple_valid_ipv6_subnet(self):
        try:
            ippol._validate_cidrs_fit_into_subnets(
                ["::/128"],
                [dict(id=1, cidr="::/96"),
                 dict(id=2, cidr="::/64")])
        except Exception:
            self.fail("Should not have failed")

    def test_normal_cidr_and_invalid_subnet(self):
        with self.assertRaises(n_exc.BadRequest):
            ippol._validate_cidrs_fit_into_subnets(
                ["192.168.0.100/32"],
                [dict(id=1, cidr="10.10.10.0/24")])

    def test_normal_ipv6_cidr_and_invalid_ipv6_subnet(self):
        with self.assertRaises(n_exc.BadRequest):
            ippol._validate_cidrs_fit_into_subnets(
                ["::/64"], [dict(id=1, cidr="::/96")])

    def test_normal_cidr_and_one_invalid_and_one_valid_subnet(self):
        with self.assertRaises(n_exc.BadRequest):
            ippol._validate_cidrs_fit_into_subnets(
                ["192.168.0.100/32"],
                [dict(id=1, cidr="10.10.10.0/24"),
                 dict(id=1, cidr="192.168.0.0/24")])

    def test_normal_ipv6_cidr_and_one_invalid_and_one_valid_ipv6_subnet(self):
        with self.assertRaises(n_exc.BadRequest):
            ippol._validate_cidrs_fit_into_subnets(
                ["::/127"],
                [dict(id=1, cidr="::/96"),
                 dict(id=1, cidr="::/128")])


class TestQuarkEnsureDefaultPolicy(test_base.TestBase):
    def test_no_cidrs_no_subnets(self):
        cidrs = []
        subnets = []
        self.assertIsNone(ippol.ensure_default_policy(cidrs, subnets))
        self.assertEqual(cidrs, [])
        self.assertEqual(subnets, [])

    def test_no_cidrs_v4(self):
        cidrs = []
        subnets = [dict(cidr="192.168.10.1/24")]
        self.assertIsNone(ippol.ensure_default_policy(cidrs, subnets))
        self.assertEqual(cidrs, ["192.168.10.0/32", "192.168.10.255/32"])
        self.assertEqual(subnets, [dict(cidr="192.168.10.1/24")])

    def test_no_subnets_v4(self):
        cidrs = ["192.168.10.0/32", "192.168.10.255/32"]
        subnets = []
        self.assertIsNone(ippol.ensure_default_policy(cidrs, subnets))
        self.assertEqual(cidrs, ["192.168.10.0/32", "192.168.10.255/32"])
        self.assertEqual(subnets, [])

    def test_cidrs_without_default_cidrs_v4(self):
        cidrs = ["192.168.10.20/32", "192.168.10.40/32"]
        subnets = [dict(cidr="192.168.10.1/24")]
        self.assertIsNone(ippol.ensure_default_policy(cidrs, subnets))
        self.assertEqual(cidrs, ["192.168.10.20/32", "192.168.10.40/32",
                                 "192.168.10.0/32", "192.168.10.255/32"])
        self.assertEqual(subnets, [dict(cidr="192.168.10.1/24")])

    def test_cidrs_with_default_cidrs_v4(self):
        cidrs = ["192.168.10.0/32", "192.168.10.255/32"]
        subnets = [dict(cidr="192.168.10.1/24")]
        self.assertIsNone(ippol.ensure_default_policy(cidrs, subnets))
        self.assertEqual(cidrs, ["192.168.10.0/32", "192.168.10.255/32"])
        self.assertEqual(subnets, [dict(cidr="192.168.10.1/24")])

    def test_no_cidrs_v6(self):
        cidrs = []
        subnets = [dict(cidr="::/64")]
        self.assertIsNone(ippol.ensure_default_policy(cidrs, subnets))
        self.assertEqual(cidrs, ["::/128", "::ffff:ffff:ffff:ffff/128"])
        self.assertEqual(subnets, [dict(cidr="::/64")])

    def test_no_subnets_v6(self):
        cidrs = ["::/128", "::ffff:ffff:ffff:ffff/128"]
        subnets = []
        self.assertIsNone(ippol.ensure_default_policy(cidrs, subnets))
        self.assertEqual(cidrs, ["::/128", "::ffff:ffff:ffff:ffff/128"])
        self.assertEqual(subnets, [])

    def test_cidrs_without_default_cidrs_v6(self):
        cidrs = ["::10/128", "::20/128"]
        subnets = [dict(cidr="::/64")]
        self.assertIsNone(ippol.ensure_default_policy(cidrs, subnets))
        self.assertEqual(cidrs, ["::10/128", "::20/128",
                                 "::/128", "::ffff:ffff:ffff:ffff/128"])
        self.assertEqual(subnets, [dict(cidr="::/64")])

    def test_cidrs_with_default_cidrs_v6(self):
        cidrs = ["::/128", "::ffff:ffff:ffff:ffff/128"]
        subnets = [dict(cidr="::/64")]
        self.assertIsNone(ippol.ensure_default_policy(cidrs, subnets))
        self.assertEqual(cidrs, ["::/128", "::ffff:ffff:ffff:ffff/128"])
        self.assertEqual(subnets, [dict(cidr="::/64")])

    def test_no_duplicates_in_result_when_called_twice(self):
        cidrs = ["192.168.10.10/32"]
        subnets = [dict(cidr="192.168.10.0/24")]
        self.assertIsNone(ippol.ensure_default_policy(cidrs, subnets))
        self.assertEqual(cidrs, ["192.168.10.10/32", "192.168.10.0/32",
                                 "192.168.10.255/32"])
        cidrs2 = ["192.168.10.10/32"]
        self.assertIsNone(ippol.ensure_default_policy(cidrs2, subnets))
        self.assertEqual(cidrs, ["192.168.10.10/32", "192.168.10.0/32",
                                 "192.168.10.255/32"])
        self.assertEqual(subnets, [dict(cidr="192.168.10.0/24")])



import mock

from neutron_lib import exceptions as n_exc
from quark.tests import test_base
from quark import utils
import webob


class TestRetryLoop(test_base.TestBase):
    def test_retry_3_times_pass(self):
        r = utils.retry_loop(3)
        c = mock.MagicMock()
        g = r(c)
        ret = g()
        c.assert_called_once_with()
        self.assertEqual(ret, c.return_value)

    def test_retry_3_times_exception(self):
        r = utils.retry_loop(3)
        c = mock.MagicMock()
        g = r(c)
        c.side_effect = ValueError()
        with self.assertRaises(ValueError):
            g()
        self.assertEqual(c.call_count, 3)

    def test_retry_3_times_exception_pass(self):
        r = utils.retry_loop(3)
        c = mock.MagicMock()
        g = r(c)
        expected_ret = mock.MagicMock()
        c.side_effect = (ValueError(), expected_ret)
        ret = g()
        self.assertEqual(c.call_count, 2)
        self.assertEqual(ret, expected_ret)


class TestExcWrapper(test_base.TestBase):

    @utils.exc_wrapper
    def raise_not_found(self):
        raise n_exc.NotFound()

    @utils.exc_wrapper
    def raise_conflict(self):
        raise n_exc.Conflict()

    @utils.exc_wrapper
    def raise_bad_request(self):
        raise n_exc.BadRequest(resource="test_bad_request", msg="testing")

    @utils.exc_wrapper
    def raise_generic_exception(self):
        raise Exception()

    @utils.exc_wrapper
    def no_raise(self):
        return ""

    def test_http_not_found(self):
        with self.assertRaises(webob.exc.HTTPNotFound):
            self.raise_not_found()

    def test_http_conflict(self):
        with self.assertRaises(webob.exc.HTTPConflict):
            self.raise_conflict()

    def test_http_bad_request(self):
        with self.assertRaises(webob.exc.HTTPBadRequest):
            self.raise_bad_request()

    def test_http_server_error(self):
        with self.assertRaises(webob.exc.HTTPInternalServerError):
            self.raise_generic_exception()

    def test_no_raise(self):
        self.assertEqual("", self.no_raise())



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
#  under the License.

import aiclib

import contextlib

import mock

import quark.db.models
import quark.drivers.optimized_nvp_driver
import quark.tests.test_nvp_driver as test_nvp_driver


class TestOptimizedNVPDriver(test_nvp_driver.TestNVPDriver):
    def setUp(self):
        if not hasattr(self, 'driver'):
            self.driver = (
                quark.drivers.optimized_nvp_driver.OptimizedNVPDriver())
        super(TestOptimizedNVPDriver, self).setUp()
        self.d_pkg = "quark.drivers.optimized_nvp_driver.OptimizedNVPDriver"
        self.context.session.add = mock.Mock(return_value=None)
        self.net_id = "12345678-1234-0000-1234-123412341234"

    def _create_lswitch_mock(self):
        lswitch = mock.Mock(id=self.lswitch_uuid, port_count=1,
                            network_id=self.net_id)
        return lswitch

    def _create_lport_mock(self, port_count):
        lport = mock.Mock()
        lport.id = self.lport_uuid
        lport.switch.id = self.lswitch_uuid
        lport.switch.port_count = port_count
        return lport


class TestOptimizedNVPDriverDeleteNetwork(TestOptimizedNVPDriver):
    '''Need to ensure that network of X switches deletes X switches.'''
    @contextlib.contextmanager
    def _stubs(self, switch_count=1):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
            mock.patch("%s._lswitch_select_by_nvp_id" % self.d_pkg),
            mock.patch("%s._lswitches_for_network" % self.d_pkg),
        ) as (conn, select_switch, get_switches):
            connection = self._create_connection()
            switch = self._create_lswitch_mock()
            conn.return_value = connection
            select_switch.return_value = switch
            get_switches.return_value = [switch] * switch_count
            self.context.session.delete = mock.Mock(return_value=None)
            yield connection, self.context.session.delete

    def test_delete_network_no_switches(self):
        '''Testing that X switches deleted with X switches; X = 0.'''
        switch_count = 0
        with self._stubs(
                switch_count=switch_count
        ) as (connection, context_delete):
            self.driver.delete_network(self.context, self.net_id)
            self.assertEqual(switch_count,
                             connection.lswitch().delete.call_count)
            self.assertEqual(switch_count, context_delete.call_count)

    def test_delete_network_single_switch(self):
        '''Testing that X switches deleted with X switches; X = 1.'''
        switch_count = 1
        with self._stubs(
                switch_count=switch_count
        ) as (connection, context_delete):
            self.driver.delete_network(self.context, self.net_id)
            self.assertEqual(switch_count,
                             connection.lswitch().delete.call_count)
            self.assertEqual(switch_count, context_delete.call_count)

    def test_delete_network_multi_switch(self):
        '''Testing that X switches deleted with X switches; X > 1.'''
        switch_count = 3
        with self._stubs(
                switch_count=switch_count
        ) as (connection, context_delete):
            self.driver.delete_network(self.context, self.net_id)
            self.assertEqual(switch_count,
                             connection.lswitch().delete.call_count)
            self.assertEqual(switch_count, context_delete.call_count)


class TestOptimizedNVPDriverDeleteNetworkWithExceptions(
        TestOptimizedNVPDriver):
    '''Test for storage of orphaned nets from NVP or our quark db tables.'''

    @contextlib.contextmanager
    def _stubs(self, switch_count=1, error_code=500):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
            mock.patch("%s._lswitch_select_by_nvp_id" % self.d_pkg),
            mock.patch("%s._lswitches_for_network" % self.d_pkg),
            mock.patch("%s._lswitch_delete" % self.d_pkg)
        ) as (conn, select_switch, get_switches, delete_switch):
            connection = self._create_connection()
            switch = self._create_lswitch_mock()
            conn.return_value = connection
            select_switch.return_value = switch
            get_switches.return_value = [switch] * switch_count
            delete_switch.side_effect = aiclib.core.AICException(
                error_code, 'foo')
            self.context.session.delete = mock.Mock(return_value=None)
            self.context.session.add = mock.Mock(return_value=None)
            yield (connection, self.context.session.delete,
                   self.context.session.add, delete_switch)

    def test_delete_network_with_404_aicexception(self):
        with self._stubs(error_code=404) as (
                connection, session_delete_network_from_db,
                session_add_orphaned_network_to_db, nvp_delete_switch):
            self.driver.delete_network(self.context, self.net_id)
            self.assertEqual(1, nvp_delete_switch.call_count)
            self.assertEqual(1, session_delete_network_from_db.call_count)
            self.assertEqual(0, session_add_orphaned_network_to_db.call_count)

    def test_delete_network_with_non_404_aicexception(self):
        with self._stubs() as (
                connection, session_delete_network_from_db,
                session_add_orphaned_network_to_db, nvp_delete_switch):
            self.driver.delete_network(self.context, self.net_id)
            self.assertEqual(1, nvp_delete_switch.call_count)
            self.assertEqual(1, session_delete_network_from_db.call_count)
            self.assertEqual(1, session_add_orphaned_network_to_db.call_count)


class TestOptimizedNVPDriverDeletePortMultiSwitch(TestOptimizedNVPDriver):
    '''Test for 0 ports on the switch, and the switch not last in network.

    Need to test if ports on switch = 0 and delete the switch if it is not
    the last switch on the network.
    '''

    @contextlib.contextmanager
    def _stubs(self, port_count=2, exception=None):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
            mock.patch("%s._lport_select_by_id" % self.d_pkg),
            mock.patch("%s._lswitch_select_by_nvp_id" % self.d_pkg),
            mock.patch("%s._lswitches_for_network" % self.d_pkg),
            mock.patch("%s._lport_delete" % self.d_pkg),
        ) as (conn, select_port, select_switch,
              two_switch, port_delete):
            connection = self._create_connection()
            port = self._create_lport_mock(port_count)
            switch = self._create_lswitch_mock()
            conn.return_value = connection
            select_port.return_value = port
            select_switch.return_value = switch
            two_switch.return_value = [switch, switch]
            self.context.session.delete = mock.Mock(return_value=None)
            if exception:
                port_delete.side_effect = exception
            yield (connection, self.context.session.delete,
                   self.context.session.add, port_delete)

    def test_delete_ports_not_empty(self):
        '''Ensure that the switch is not deleted if ports exist.'''
        with self._stubs() as (
                connection, context_delete, context_add, port_delete):
            self.driver.delete_port(self.context, self.port_id)
            self.assertEqual(1, context_delete.call_count)
            self.assertTrue(port_delete.called)
            self.assertFalse(connection.lswitch().delete.called)

    def test_delete_ports_is_empty(self):
        '''Ensure that the switch is deleted if empty and not last switch.'''
        with self._stubs(port_count=1) as (
                connection, context_delete, context_add, port_delete):
            self.driver.delete_port(self.context, self.port_id)
            self.assertEqual(2, context_delete.call_count)
            self.assertTrue(port_delete.called)
            self.assertTrue(connection.lswitch().delete.called)

    def test_delete_ports_with_exception(self):
        '''Ensure that exception is handled/logged.'''
        e = Exception('foo')
        with self._stubs(exception=e) as (
                connection, context_delete, context_add, port_delete):
            try:
                with self.assertRaises(type(e)):
                    self.driver.delete_port(self.context, self.port_id)
                self.fail("AssertionError should have been raised.")
            except AssertionError as ae:
                self.assertEqual(ae.args[0], "Exception not raised")
                self.assertEqual(1, context_delete.call_count)
                self.assertEqual(0, context_add.call_count)
                self.assertFalse(connection.lswitch_port().delete.called)
                self.assertFalse(connection.lswitch().delete.called)

    def test_delete_ports_with_404_aicexception(self):
        '''Ensure that exception is handled/logged.'''
        e = aiclib.core.AICException(404, 'foo')
        with self._stubs(exception=e) as (
                connection, context_delete, context_add, port_delete):
            try:
                with self.assertRaises(type(e)):
                    self.driver.delete_port(self.context, self.port_id)
                self.fail("AssertionError should have been raised.")
            except AssertionError as ae:
                self.assertEqual(ae.args[0], "AICException not raised")
                self.assertEqual(1, context_delete.call_count)
                self.assertEqual(0, context_add.call_count)
                self.assertFalse(connection.lswitch_port().delete.called)
                self.assertFalse(connection.lswitch().delete.called)

    def test_delete_ports_with_500_aicexception(self):
        '''Ensure that exception is handled/logged.'''
        e = aiclib.core.AICException(500, 'foo')
        with self._stubs(exception=e) as (
                connection, context_delete, context_add, port_delete):
            try:
                with self.assertRaises(type(e)):
                    self.driver.delete_port(self.context, self.port_id)
                self.fail("AssertionError should have been raised.")
            except AssertionError as ae:
                self.assertEqual(ae.args[0], "AICException not raised")
                self.assertEqual(1, context_delete.call_count)
                self.assertEqual(1, context_add.call_count)
                self.assertFalse(connection.lswitch_port().delete.called)
                self.assertFalse(connection.lswitch().delete.called)


class TestOptimizedNVPDriverDeletePortSingleSwitch(TestOptimizedNVPDriver):
    '''If ports on switch = 0, delete switch unless last on the network.'''

    @contextlib.contextmanager
    def _stubs(self, port_count=1):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
            mock.patch("%s._lport_select_by_id" % self.d_pkg),
            mock.patch("%s._lswitch_select_by_nvp_id" % self.d_pkg),
            mock.patch("%s._lswitches_for_network" % self.d_pkg),
        ) as (conn, select_port, select_switch, one_switch):
            connection = self._create_connection()
            switch = self._create_lswitch_mock()
            conn.return_value = connection
            select_port.return_value = None
            select_switch.return_value = switch
            one_switch.return_value = [switch]
            self.context.session.delete = mock.Mock(return_value=None)
            yield connection, self.context.session.delete

    def test_delete_ports_is_empty(self):
        '''Ensure that the switch is not deleted if it is the last.'''
        with self._stubs(port_count=1) as (connection, context_delete):
            self.driver.delete_port(self.context, self.port_id)
            self.assertEqual(0, context_delete.call_count)
            self.assertFalse(connection.lswitch_port().delete.called)


class TestOptimizedNVPDriverDeletePortMissing(TestOptimizedNVPDriver):
    '''If ports on switch = 0, delete switch unless last on the network.'''

    @contextlib.contextmanager
    def _stubs(self, port_count=2):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
            mock.patch("%s._lport_select_by_id" % self.d_pkg),
            mock.patch("%s._lswitch_select_by_nvp_id" % self.d_pkg),
            mock.patch("%s._lswitches_for_network" % self.d_pkg),
        ) as (conn, select_port, select_switch, one_switch):
            connection = self._create_connection()
            port = self._create_lport_mock(port_count)
            switch = self._create_lswitch_mock()
            conn.return_value = connection
            select_port.return_value = port
            select_switch.return_value = switch
            one_switch.return_value = [switch]
            self.context.session.delete = mock.Mock(return_value=None)
            yield connection, self.context.session.delete

    def test_delete_ports_is_empty(self):
        '''Ensure that the switch is not deleted if it is the last.'''
        with self._stubs(port_count=1) as (connection, context_delete):
            self.driver.delete_port(self.context, self.port_id)
            self.assertEqual(1, context_delete.call_count)
            self.assertTrue(connection.lswitch_port().delete.called)
            self.assertFalse(connection.lswitch().delete.called)


class TestOptimizedNVPDriverCreatePort(TestOptimizedNVPDriver):
    '''In no case should the optimized driver query for an lswitch.'''

    @contextlib.contextmanager
    def _stubs(self, has_lswitch=True, maxed_ports=False):
        with contextlib.nested(
            mock.patch("%s._connection" % self.d_pkg),
            mock.patch("%s._lswitch_select_free" % self.d_pkg),
            mock.patch("%s._lswitch_select_first" % self.d_pkg),
            mock.patch("%s._lswitch_select_by_nvp_id" % self.d_pkg),
            mock.patch("%s._lswitch_create_optimized" % self.d_pkg),
            mock.patch("%s._get_network_details" % self.d_pkg)
        ) as (conn, select_free, select_first,
              select_by_id, create_opt, get_net_dets):
            connection = self._create_connection()
            conn.return_value = connection
            if has_lswitch:
                select_first.return_value = mock.Mock(nvp_id=self.lswitch_uuid)
            if not has_lswitch:
                select_first.return_value = None
                select_free.return_value = None
            elif not maxed_ports:
                select_free.return_value = self._create_lswitch_mock()
            else:
                select_free.return_value = None

            select_by_id.return_value = self._create_lswitch_mock()
            get_net_dets.return_value = dict(foo=3)
            yield connection, create_opt

    def test_select_ipam_strategy(self):
        strategy = self.driver.select_ipam_strategy(1, "ANY")
        self.assertEqual(strategy, "ANY")

    def test_create_port_and_maxed_switch_spanning(self):
        '''Testing to ensure a switch is made when maxed.'''
        with self._stubs(maxed_ports=True) as (
                connection, create_opt):
            self.driver.limits['max_ports_per_switch'] = self.max_spanning
            port = self.driver.create_port(self.context, self.net_id,
                                           self.port_id)
            self.assertTrue("uuid" in port)
            self.assertTrue(connection.lswitch().create.called)
            self.assertTrue(connection.lswitch_port().create.called)
            self.assertTrue(create_opt.called)
            self.assertFalse(connection.lswitch().query.called)
            status_args, kwargs = self.context.session.add.call_args
            status_args, kwargs = (
                connection.lswitch_port().admin_status_enabled.call_args)
            self.assertTrue(True in status_args)

    def test_create_port_and_create_switch_spanning(self):
        '''Testing to ensure a switch is made when no switch available.'''
        with self._stubs(has_lswitch=False) as (connection, create_opt):
            self.driver.max_ports_per_switch = self.max_spanning
            port = self.driver.create_port(self.context, self.net_id,
                                           self.port_id)
            self.assertTrue("uuid" in port)
            self.assertTrue(connection.lswitch().create.called)
            self.assertTrue(connection.lswitch_port().create.called)
            self.assertTrue(create_opt.called)
            self.assertFalse(connection.lswitch().query.called)
            status_args, kwargs = self.context.session.add.call_args
            status_args, kwargs = (
                connection.lswitch_port().admin_status_enabled.call_args)
            self.assertTrue(True in status_args)

    def test_create_port_and_no_create_switch_spanning(self):
        '''Testing to ensure a switch is not made when max ports not met.'''
        with self._stubs() as (connection, create_opt):
            self.driver.max_ports_per_switch = self.max_spanning
            port = self.driver.create_port(self.context, self.net_id,
                                           self.port_id)
            self.assertTrue("uuid" in port)
            self.assertFalse(connection.lswitch().create.called)
            self.assertTrue(connection.lswitch_port().create.called)
            self.assertFalse(create_opt.called)
            self.assertFalse(connection.lswitch().query.called)
            status_args, kwargs = self.context.session.add.call_args
            status_args, kwargs = (
                connection.lswitch_port().admin_status_enabled.call_args)
            self.assertTrue(True in status_args)

    def test_create_disabled_port_and_no_create_switch_spanning(self):
        '''Testing to ensure a port is made and disabled.'''
        with self._stubs() as (connection, create_opt):
            self.driver.max_ports_per_switch = self.max_spanning
            port = self.driver.create_port(self.context, self.net_id,
                                           self.port_id, False)
            self.assertTrue("uuid" in port)
            self.assertFalse(connection.lswitch().create.called)
            self.assertTrue(connection.lswitch_port().create.called)
            self.assertFalse(create_opt.called)
            self.assertFalse(connection.lswitch().query.called)
            status_args, kwargs = self.context.session.add.call_args
            status_args, kwargs = (
                connection.lswitch_port().admin_status_enabled.call_args)
            self.assertTrue(False in status_args)

    def test_create_port_and_create_switch(self):
        '''Testing to ensure a switch is made when no switch available.'''
        with self._stubs(has_lswitch=False) as (connection, create_opt):
            port = self.driver.create_port(self.context, self.net_id,
                                           self.port_id)
            self.assertTrue("uuid" in port)
            self.assertTrue(connection.lswitch().create.called)
            self.assertTrue(connection.lswitch_port().create.called)
            self.assertTrue(create_opt.called)
            self.assertFalse(connection.lswitch().query.called)
            status_args, kwargs = self.context.session.add.call_args
            status_args, kwargs = (
                connection.lswitch_port().admin_status_enabled.call_args)
            self.assertTrue(True in status_args)

    def test_create_port_and_no_create_switch(self):
        '''Testing to ensure a switch is not made when available.'''
        with self._stubs() as (connection, create_opt):
            port = self.driver.create_port(self.context, self.net_id,
                                           self.port_id)
            self.assertTrue("uuid" in port)
            self.assertFalse(connection.lswitch().create.called)
            self.assertTrue(connection.lswitch_port().create.called)
            self.assertFalse(create_opt.called)
            self.assertFalse(connection.lswitch().query.called)
            status_args, kwargs = self.context.session.add.call_args
            status_args, kwargs = (
                connection.lswitch_port().admin_status_enabled.call_args)
            self.assertTrue(True in status_args)

    def test_create_disabled_port_and_no_create_switch(self):
        '''Testing to ensure a port is made and disabled.'''
        with self._stubs() as (connection, create_opt):
            port = self.driver.create_port(self.context, self.net_id,
                                           self.port_id, False)
            self.assertTrue("uuid" in port)
            self.assertFalse(connection.lswitch().create.called)
            self.assertTrue(connection.lswitch_port().create.called)
            self.assertFalse(create_opt.called)
            self.assertFalse(connection.lswitch().query.called)
            status_args, kwargs = self.context.session.add.call_args
            status_args, kwargs = (
                connection.lswitch_port().admin_status_enabled.call_args)
            self.assertTrue(False in status_args)


class TestOptimizedNVPDriverUpdatePort(TestOptimizedNVPDriver):
    def test_update_port(self):
        mod_path = "quark.drivers.%s"
        op_path = "optimized_nvp_driver.OptimizedNVPDriver"
        lport_path = "%s._lport_select_by_id" % op_path
        with contextlib.nested(
            mock.patch(mod_path % "nvp_driver.NVPDriver.update_port"),
            mock.patch(mod_path % lport_path),
        ) as (update_port, port_find):
            ret_port = quark.drivers.optimized_nvp_driver.LSwitchPort()
            port_find.return_value = ret_port
            update_port.return_value = dict(switch_id=2)
            self.driver.update_port(self.context, 1)
            self.assertEqual(ret_port.switch_id, 2)


class TestCreateSecurityGroups(TestOptimizedNVPDriver):
    def test_create_security_group(self):
        with mock.patch("%s._connection" % self.d_pkg):
            self.driver.create_security_group(self.context, "newgroup")
            self.assertTrue(self.context.session.add.called)


class TestDeleteSecurityGroups(TestOptimizedNVPDriver):
    def test_delete_security_group(self):
        mod_path = "quark.drivers.nvp_driver.NVPDriver"
        with contextlib.nested(
                mock.patch("%s._connection" % self.d_pkg),
                mock.patch("%s._query_security_group" % self.d_pkg),
                mock.patch("%s.delete_security_group" % mod_path)):

            session_delete = self.context.session.delete
            self.context.session.delete = mock.Mock(return_value=None)
            self.driver.delete_security_group(self.context, 1)
            self.assertTrue(self.context.session.delete.called)
            self.context.session.delete = session_delete


class TestSecurityGroupRules(TestOptimizedNVPDriver):
    @contextlib.contextmanager
    def _stubs(self, rules=None):
        rules = rules or []
        with contextlib.nested(
                mock.patch("%s._connection" % self.d_pkg),
                mock.patch("%s._query_security_group" % self.d_pkg),
                mock.patch("%s._check_rule_count_per_port" % self.d_pkg),
        ) as (conn, query_sec_group, rule_count):
            query_sec_group.return_value = (quark.drivers.optimized_nvp_driver.
                                            SecurityProfile())
            connection = self._create_connection()
            rule_count.return_value = 1
            connection.securityprofile = self._create_security_profile()
            connection.securityrule = self._create_security_rule()
            connection.lswitch_port().query.return_value = (
                self._create_lport_query(1, [self.profile_id]))
            conn.return_value = connection

            old_query = self.context.session.query
            sec_group = quark.db.models.SecurityGroup()
            for rule in rules:
                rule_mod = quark.db.models.SecurityGroupRule()
                rule_mod.update(rule)
                sec_group.rules.append(rule_mod)
            first_mock = mock.Mock()
            filter_mock = mock.Mock()
            self.context.session.query = mock.Mock(return_value=filter_mock)
            filter_mock.filter.return_value = first_mock
            first_mock.first.return_value = sec_group

            yield connection
            self.context.session.query = old_query

    def test_security_rule_create_no_rules(self):
        with self._stubs() as connection:
            self.driver.create_security_group_rule(
                self.context, 1,
                {'ethertype': 'IPv4', 'direction': 'ingress'})
            connection.securityprofile().assert_has_calls([
                mock.call.port_ingress_rules([{'ethertype': 'IPv4'}]),
                mock.call.update(),
            ], any_order=True)

    def test_security_rule_create(self):
        with self._stubs(rules=[{"direction": "ingress"}]) as connection:
            self.driver.create_security_group_rule(
                self.context, 1,
                {'ethertype': 'IPv4', 'direction': 'ingress'})
            connection.securityprofile().assert_has_calls([
                mock.call.port_ingress_rules([{}, {'ethertype': 'IPv4'}]),
                mock.call.update(),
            ], any_order=True)


class TestCreateLswitchOptimized(TestOptimizedNVPDriver):
    def test_create_lswitch_optimized(self):
        self.driver._lswitch_create_optimized(self.context, "public", 1, 1)
        self.assertTrue(self.context.session.add.called)


class TestGetNetworkDetails(TestOptimizedNVPDriver):
    @contextlib.contextmanager
    def _stubs(self, switch):
        with mock.patch("%s._lswitch_select_first" % self.d_pkg) as lselect:
            switch_model = quark.drivers.optimized_nvp_driver.LSwitch()
            switch_model.update(switch)
            lselect.return_value = switch_model
            yield

    def test_get_network_details(self):
        switch = {"network_id": 2, "transport_zone": 1,
                  "transport_connector": "bridge", "segment_id": 3,
                  "display_name": "public"}
        expected = {"network_name": "public", "phys_net": 1,
                    "phys_type": "bridge", "segment_id": 3}
        with self._stubs(switch):
            switches = []
            details = self.driver._get_network_details(self.context, 1,
                                                       switches)
            for key in expected.keys():
                self.assertEqual(details[key], expected[key])


class TestQueryMethods(TestOptimizedNVPDriver):
    """These tests provide coverage on the query helpers.

    No serious assertions are made, as there's no sense in testing that
    sqlalchemy does in fact do what it's supposed to do.
    """

    @contextlib.contextmanager
    def _stubs(self):
        old_query = self.context.session.query
        self.context.session.query = mock.Mock()
        query_return = mock.MagicMock()
        self.context.session.query.return_value = query_return
        yield query_return
        self.context.session.query = old_query

    def test_lport_select_by_id(self):
        with self._stubs() as query_return:
            self.driver._lport_select_by_id(self.context, 1)
            self.assertTrue(query_return.filter.called)

    def test_lswitch_select_by_nvp_id(self):
        with self._stubs() as query_return:
            self.driver._lswitch_select_by_nvp_id(self.context, 1)
            self.assertTrue(query_return.filter.called)

    def test_lswitch_select_first(self):
        with self._stubs() as query_return:
            self.driver._lswitch_select_first(self.context, 1)
            self.assertTrue(query_return.filter.called)

            # NOTE(mdietz): This is probably pretty brittle, but I didn't want
            #               this patch going in without a test
            bin_expr = query_return.filter.__dict__["_mock_call_args"][0][0]
            children = bin_expr.get_children()
            self.assertEqual(str(children[0]),
                             "quark_nvp_driver_lswitch.network_id")

    def test_lswitch_select_free(self):
        with self._stubs() as query_return:
            self.driver._lswitch_select_free(self.context, 1)
            self.assertTrue(query_return.filter.called)

    def test_lswitches_for_network(self):
        with self._stubs() as query_return:
            self.driver._lswitches_for_network(self.context, 1)
            self.assertTrue(query_return.filter.called)

    def test_get_lswitch_ids_for_network(self):
        with self._stubs() as query_return:
            query_result = query_return.filter.return_value.all
            query_result.return_value = [{"nvp_id": "foo"}]
            ids = self.driver.get_lswitch_ids_for_network(self.context, 1)
            self.assertTrue(query_return.filter.called)
            self.assertEqual(ids, ["foo"])

    def test_lswitch_from_port(self):
        with self._stubs() as query_return:
            self.driver._lswitch_from_port(self.context, 1)
            self.assertTrue(query_return.filter.called)

    def test_query_security_group(self):
        with self._stubs() as query_return:
            self.driver._query_security_group(self.context, 1)
            self.assertTrue(query_return.filter.called)



# Copyright 2013 Rackspace Hosting Inc.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for# the specific language governing permissions and limitations
#  under the License.

import mock
import netaddr
from neutron.api.v2 import attributes as neutron_attrs
from neutron_lib import exceptions as n_exc
from oslo_config import cfg

import contextlib

from quark.db import api as db_api
import quark.ipam
# import below necessary if file run by itself
from quark import plugin  # noqa
import quark.plugin_modules.ip_policies as policy_api
import quark.plugin_modules.networks as network_api
import quark.plugin_modules.subnets as subnet_api
from quark.tests.functional.base import BaseFunctionalTest

CONF = cfg.CONF


class QuarkGetSubnets(BaseFunctionalTest):
    @contextlib.contextmanager
    def _stubs(self, network, subnet):
        self.ipam = quark.ipam.QuarkIpamANY()
        with self.context.session.begin():
            net_mod = db_api.network_create(self.context, **network)
            subnet["network"] = net_mod
            sub1 = db_api.subnet_create(self.context, **subnet)
            subnet["id"] = 2
            sub2 = db_api.subnet_create(self.context, do_not_use=True,
                                        **subnet)
        yield net_mod, sub1, sub2

    def test_get_subnet_do_not_use_not_returned(self):
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        subnet = dict(id=1, ip_version=4, next_auto_assign_ip=2,
                      cidr="0.0.0.0/24", first_ip=0, last_ip=255,
                      ip_policy=None, tenant_id="fake")
        with self._stubs(network, subnet) as (net, sub1, sub2):
            subnets = db_api.subnet_find_ordered_by_most_full(self.context,
                                                              net["id"]).all()
            self.assertEqual(len(subnets), 1)
            self.assertEqual(subnets[0][0]["id"], "1")


class QuarkGetSubnetsFromPlugin(BaseFunctionalTest):
    @contextlib.contextmanager
    def _stubs(self, network, subnet):
        self.ipam = quark.ipam.QuarkIpamANY()
        with contextlib.nested(mock.patch("neutron.common.rpc.get_notifier")):
            net = network_api.create_network(self.context, network)
            subnet['subnet']['network_id'] = net['id']
            sub1 = subnet_api.create_subnet(self.context, subnet)
            yield net, sub1

    def test_toggle_ip_policy_id_from_subnet_view(self):
        cidr = "192.168.1.0/24"
        ip_network = netaddr.IPNetwork(cidr)
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        network = {"network": network}
        subnet = dict(id=1, ip_version=4, next_auto_assign_ip=2,
                      cidr=cidr, first_ip=ip_network.first,
                      last_ip=ip_network.last, ip_policy=None,
                      tenant_id="fake")
        subnet = {"subnet": subnet}
        original = cfg.CONF.QUARK.show_subnet_ip_policy_id
        with self._stubs(network, subnet) as (net, sub1):
            cfg.CONF.set_override('show_subnet_ip_policy_id', True, "QUARK")
            subnet = subnet_api.get_subnet(self.context, 1)
            self.assertTrue('ip_policy_id' in subnet)
            cfg.CONF.set_override('show_subnet_ip_policy_id', False, "QUARK")
            subnet = subnet_api.get_subnet(self.context, 1)
            self.assertFalse('ip_policy_id' in subnet)
        cfg.CONF.set_override('show_subnet_ip_policy_id', original, "QUARK")


class QuarkCreateSubnets(BaseFunctionalTest):
    @contextlib.contextmanager
    def _stubs(self, network, subnet):
        self.ipam = quark.ipam.QuarkIpamANY()
        with contextlib.nested(mock.patch("neutron.common.rpc.get_notifier")):
            net = network_api.create_network(self.context, network)
            subnet['subnet']['network_id'] = net['id']
            sub1 = subnet_api.create_subnet(self.context, subnet)
            yield net, sub1

    def test_create_allocation_pools_over_quota_fail(self):
        original_pool_quota = cfg.CONF.QUOTAS.quota_alloc_pools_per_subnet
        cidr = "1.1.1.0/8"
        ip_network = netaddr.IPNetwork(cidr)
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        network = {"network": network}
        pools = [{"start": "1.0.1.2", "end": "1.0.2.0"},
                 {"start": "1.0.2.2", "end": "1.0.3.0"}]
        subnet = dict(id=1, ip_version=4, next_auto_assign_ip=2,
                      cidr=cidr, first_ip=ip_network.first,
                      last_ip=ip_network.last, ip_policy=None,
                      tenant_id="fake", allocation_pools=pools)
        subnet = {"subnet": subnet}
        with self.assertRaises(n_exc.OverQuota):
            cfg.CONF.set_override('quota_alloc_pools_per_subnet', 1, "QUOTAS")
            with self._stubs(network, subnet) as (net, sub):
                self.assertTrue(sub)
        cfg.CONF.set_override('quota_alloc_pools_per_subnet',
                              original_pool_quota, "QUOTAS")

    def test_create_allocation_pools_under_quota_pass(self):
        original_pool_quota = cfg.CONF.QUOTAS.quota_alloc_pools_per_subnet
        cidr = "1.1.1.0/8"
        ip_network = netaddr.IPNetwork(cidr)
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        network = {"network": network}
        pools = [{"start": "1.0.1.2", "end": "1.0.2.0"}]
        subnet = dict(id=1, ip_version=4, next_auto_assign_ip=2,
                      cidr=cidr, first_ip=ip_network.first,
                      last_ip=ip_network.last, ip_policy=None,
                      tenant_id="fake", allocation_pools=pools)
        subnet = {"subnet": subnet}
        cfg.CONF.set_override('quota_alloc_pools_per_subnet', 1, "QUOTAS")
        with self._stubs(network, subnet) as (net, sub):
            self.assertTrue(sub)
        cfg.CONF.set_override('quota_alloc_pools_per_subnet',
                              original_pool_quota, "QUOTAS")

    def test_create_allocation_pools_empty(self):
        cidr = "192.168.1.0/24"
        ip_network = netaddr.IPNetwork(cidr)
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        network = {"network": network}
        pools = []
        subnet = dict(id=1, ip_version=4, next_auto_assign_ip=2,
                      cidr=cidr, first_ip=ip_network.first,
                      last_ip=ip_network.last, ip_policy=None,
                      tenant_id="fake", allocation_pools=pools)
        subnet = {"subnet": subnet}
        with self._stubs(network, subnet) as (net, sub1):
            self.assertEqual(sub1["allocation_pools"], [])

    def test_create_allocation_pools_none(self):
        cidr = "192.168.1.0/24"
        ip_network = netaddr.IPNetwork(cidr)
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        network = {"network": network}
        pools = None
        subnet = dict(id=1, ip_version=4, next_auto_assign_ip=2,
                      cidr=cidr, first_ip=ip_network.first,
                      last_ip=ip_network.last, ip_policy=None,
                      tenant_id="fake", allocation_pools=pools)
        subnet = {"subnet": subnet}
        with self._stubs(network, subnet) as (net, sub1):
            self.assertEqual(sub1["allocation_pools"],
                             [dict(start="192.168.1.1", end="192.168.1.254")])

    def test_create_allocation_pools_full(self):
        cidr = "192.168.1.0/24"
        ip_network = netaddr.IPNetwork(cidr)
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        network = {"network": network}
        pools = [dict(start="192.168.1.0", end="192.168.1.255")]
        subnet = dict(id=1, ip_version=4, next_auto_assign_ip=2,
                      cidr=cidr, first_ip=ip_network.first,
                      last_ip=ip_network.last, ip_policy=None,
                      tenant_id="fake", allocation_pools=pools)
        subnet = {"subnet": subnet}
        with self._stubs(network, subnet) as (net, sub1):
            self.assertEqual(sub1["allocation_pools"],
                             [dict(start="192.168.1.1", end="192.168.1.254")])

    def test_create_ipv6_subnet_with_multiple_allocation_pools(self):
        cidr = "fd00:243:e319::/64"
        ip_network = netaddr.IPNetwork(cidr)
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        network = {"network": network}
        pools = [{'start': 'fd00:243:e319::64', 'end': 'fd00:243:e319::384'},
                 {'start': 'fd00:243:e319::3e8', 'end': 'fd00:243:e319::76c'},
                 {'start': 'fd00:243:e319::7d0', 'end': 'fd00:243:e319::b54'}]
        next_auto_assign_ip = ip_network.first + 1
        subnet = dict(id=1, ip_versino=6,
                      next_auto_assign_ip=next_auto_assign_ip,
                      cidr=cidr, first_ip=ip_network.first,
                      last_ip=ip_network.last, ip_policy=None,
                      tenant_id="fake", allocation_pools=pools)
        subnet = {"subnet": subnet}
        with self._stubs(network, subnet) as (net, sub1):
            self.assertEqual(sub1["allocation_pools"], pools)

    def test_create_cidr_not_specified_raises(self):
        cidr = neutron_attrs.ATTR_NOT_SPECIFIED
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        network = {"network": network}
        subnet = dict(ip_version=4, cidr=cidr, tenant_id="fake")
        subnet = {"subnet": subnet}
        with self.assertRaises(n_exc.BadRequest):
            with self._stubs(network, subnet) as (net, sub1):
                pass


class QuarkUpdateSubnets(BaseFunctionalTest):

    def setUp(self):
        super(QuarkUpdateSubnets, self).setUp()
        self.og = CONF.QUARK.allow_allocation_pool_update
        self.og1 = CONF.QUARK.allow_allocation_pool_growth
        CONF.set_override('allow_allocation_pool_update', True, 'QUARK')
        CONF.set_override('allow_allocation_pool_growth', True, 'QUARK')

    def tearDown(self):
        super(QuarkUpdateSubnets, self).tearDown()
        CONF.set_override('allow_allocation_pool_update', self.og, 'QUARK')
        CONF.set_override('allow_allocation_pool_growth', self.og1, 'QUARK')

    @contextlib.contextmanager
    def _stubs(self, network, subnet):
        self.ipam = quark.ipam.QuarkIpamANY()
        with contextlib.nested(mock.patch("neutron.common.rpc.get_notifier")):
            net = network_api.create_network(self.context, network)
            subnet['subnet']['network_id'] = net['id']
            sub1 = subnet_api.create_subnet(self.context, subnet)
            yield net, sub1

    def test_update_allocation_pools(self):
        cidr = "192.168.1.0/24"
        ip_network = netaddr.IPNetwork(cidr)
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        network = {"network": network}
        subnet = dict(id=1, ip_version=4, next_auto_assign_ip=2,
                      cidr=cidr, first_ip=ip_network.first,
                      last_ip=ip_network.last, ip_policy=None,
                      tenant_id="fake")
        subnet = {"subnet": subnet}
        with self._stubs(network, subnet) as (net, sub1):
            subnet = subnet_api.get_subnet(self.context, 1)
            start_pools = subnet['allocation_pools']
            new_pools = [
                [dict(start='192.168.1.10', end='192.168.1.50')],
                [dict(start='192.168.1.5', end='192.168.1.25')],
                [dict(start='192.168.1.50', end='192.168.1.51')],
                [dict(start='192.168.1.50', end='192.168.1.51'),
                    dict(start='192.168.1.100', end='192.168.1.250')],
                [dict(start='192.168.1.50', end='192.168.1.51')],
                start_pools,
            ]
            prev_pool = start_pools
            for pool in new_pools:
                subnet_update = {"subnet": dict(allocation_pools=pool)}
                subnet = subnet_api.update_subnet(self.context, 1,
                                                  subnet_update)
                self.assertNotEqual(prev_pool, subnet['allocation_pools'])
                self.assertEqual(pool, subnet['allocation_pools'])
                policies = policy_api.get_ip_policies(self.context)
                self.assertEqual(1, len(policies))
                policy = policies[0]
                ip_set = netaddr.IPSet()
                for ip in policy['exclude']:
                    ip_set.add(netaddr.IPNetwork(ip))
                for extent in pool:
                    for ip in netaddr.IPRange(extent['start'], extent['end']):
                        self.assertFalse(ip in ip_set)
                prev_pool = pool

    def test_allow_allocation_pool_growth(self):
        CONF.set_override('allow_allocation_pool_growth', True, 'QUARK')
        cidr = "192.168.1.0/24"
        ip_network = netaddr.IPNetwork(cidr)
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        network = {"network": network}
        pool = [dict(start='192.168.1.15', end='192.168.1.30')]
        subnet = dict(id=1, ip_version=4, next_auto_assign_ip=2,
                      cidr=cidr, first_ip=ip_network.first,
                      last_ip=ip_network.last, ip_policy=None,
                      allocation_pools=pool, tenant_id="fake")
        subnet = {"subnet": subnet}
        with self._stubs(network, subnet) as (net, sub1):
            subnet = subnet_api.get_subnet(self.context, 1)
            start_pools = subnet['allocation_pools']
            new_pool = [dict(start='192.168.1.10', end='192.168.1.50')]

            subnet_update = {"subnet": dict(allocation_pools=new_pool)}
            subnet = subnet_api.update_subnet(self.context, 1,
                                              subnet_update)
            self.assertNotEqual(start_pools, subnet['allocation_pools'])
            self.assertEqual(new_pool, subnet['allocation_pools'])
            policies = policy_api.get_ip_policies(self.context)
            self.assertEqual(1, len(policies))
            policy = policies[0]
            ip_set = netaddr.IPSet()
            for ip in policy['exclude']:
                ip_set.add(netaddr.IPNetwork(ip))
            for extent in new_pool:
                for ip in netaddr.IPRange(extent['start'], extent['end']):
                    self.assertFalse(ip in ip_set)

            start_ip_set = netaddr.IPSet()
            for rng in start_pools:
                start_ip_set.add(netaddr.IPRange(rng['start'], rng['end']))

            new_ip_set = netaddr.IPSet()
            for rng in subnet['allocation_pools']:
                new_ip_set.add(netaddr.IPRange(rng['start'], rng['end']))

            self.assertTrue(start_ip_set | new_ip_set != start_ip_set)

    def test_do_not_allow_allocation_pool_growth(self):
        CONF.set_override('allow_allocation_pool_growth', False, 'QUARK')
        cidr = "192.168.1.0/24"
        ip_network = netaddr.IPNetwork(cidr)
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        network = {"network": network}
        pool = [dict(start='192.168.1.15', end='192.168.1.30')]
        subnet = dict(id=1, ip_version=4, next_auto_assign_ip=2,
                      cidr=cidr, first_ip=ip_network.first,
                      last_ip=ip_network.last, ip_policy=None,
                      allocation_pools=pool, tenant_id="fake")
        subnet = {"subnet": subnet}
        with self._stubs(network, subnet) as (net, sub1):
            subnet = subnet_api.get_subnet(self.context, 1)
            start_pools = subnet['allocation_pools']
            new_pool = [dict(start='192.168.1.10', end='192.168.1.50')]

            start_ip_set = netaddr.IPSet()
            for rng in start_pools:
                start_ip_set.add(netaddr.IPRange(rng['start'], rng['end']))

            new_ip_set = netaddr.IPSet()
            for rng in new_pool:
                new_ip_set.add(netaddr.IPRange(rng['start'], rng['end']))

            self.assertTrue(start_ip_set | new_ip_set != start_ip_set)

            subnet_update = {"subnet": dict(allocation_pools=new_pool)}
            with self.assertRaises(n_exc.BadRequest):
                subnet = subnet_api.update_subnet(self.context, 1,
                                                  subnet_update)

    def _test_allow_allocation_pool_identity(self, conf_flag):
        CONF.set_override('allow_allocation_pool_growth', conf_flag, 'QUARK')
        cidr = "192.168.1.0/24"
        ip_network = netaddr.IPNetwork(cidr)
        network = dict(name="public", tenant_id="fake", network_plugin="BASE")
        network = {"network": network}
        pool = [dict(start='192.168.1.15', end='192.168.1.30')]
        subnet = dict(id=1, ip_version=4, next_auto_assign_ip=2,
                      cidr=cidr, first_ip=ip_network.first,
                      last_ip=ip_network.last, ip_policy=None,
                      allocation_pools=pool, tenant_id="fake")
        subnet = {"subnet": subnet}
        with self._stubs(network, subnet) as (net, sub1):
            subnet = subnet_api.get_subnet(self.context, 1)
            start_pools = subnet['allocation_pools']
            new_pool = [dict(start='192.168.1.15', end='192.168.1.30')]

            start_ip_set = netaddr.IPSet()
            for rng in start_pools:
                start_ip_set.add(netaddr.IPRange(rng['start'], rng['end']))

            new_ip_set = netaddr.IPSet()
            for rng in new_pool:
                new_ip_set.add(netaddr.IPRange(rng['start'], rng['end']))

            self.assertTrue(start_ip_set == new_ip_set)

            subnet_update = {"subnet": dict(allocation_pools=new_pool)}
            subnet = subnet_api.update_subnet(self.context, 1, subnet_update)
            self.assertEqual(start_pools, subnet['allocation_pools'])
            self.assertEqual(new_pool, subnet['allocation_pools'])

    def test_allow_allocation_pool_identity_when_growth_false(self):
        self._test_allow_allocation_pool_identity(False)

    def test_allow_allocation_pool_identity_when_growth_true(self):
        self._test_allow_allocation_pool_identity(True)
